"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

Create a blank rectangle matplotlib patch

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
import matplotlib as mpl


def blank_legend_key():
    """
    Show only text in matplotlib legenda, no key
    """
    return mpl.patches.Rectangle((0, 0), 0, 0, fill=False,
                                 edgecolor='none', visible=False)



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

Implements a lite interface with the xarray::DataArray, should work for all
2-D detector images, sequences of detector measurements and trend data.

Data values
-----------
* All floating dataset are converted to Python-compatible floating-point number.

Dimensions and Coordinates
--------------------------
* The functions in this module should work with netCDF4 and HDF5 files.
* In a HDF5 file the 'coordinates' of a dataset can be defined using
 dimension scales.
* In a netCDF4 file this is required: all variables have dimensions, which
 can have coordinates. But under the hood also netCDF4 uses dimension scales.
* The xarray DataArray structure will have as dimensions, the names of the
 dimension scales and as coordinates the names and data of the dimensions
 scales, except when the data only contains zero's.
* The default dimensions of an image are 'row' and 'column' with evenly spaced
 values created with np.arange(len(dim), dtype=uint).

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
from pathlib import PurePath

import numpy as np
import xarray as xr


def __get_attrs(dset, field: str) -> dict:
    """
    Return attributes of the HDF5 dataset

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the attributes are read
    field : str
       Name of field in compound dataset

    Returns
    -------
    dict with numpy arrays
    """
    _field = None
    if field is not None:
        try:
            _field = {'name': field,
                      'oneof': len(dset.dtype.names),
                      'index': dset.dtype.names.index(field)}
        except Exception as exc:
            raise RuntimeError(
                f'field {field} not found in dataset {dset.name}') from exc
        # print('_field ', _field)

    attrs = {}
    for key in dset.attrs:
        if key in ('CLASS', 'DIMENSION_LIST', 'NAME', 'REFERENCE_LIST',
                   '_Netcdf4Dimid', '_Netcdf4Coordinates'):
            continue

        attr_value = dset.attrs[key]
        # print('# ----- ', key, type(attr_value), attr_value)
        if isinstance(attr_value, np.ndarray):
            if len(attr_value) == 1:
                attr_value = attr_value[0]
                # print('# ----- ', key, type(attr_value), attr_value)
            elif _field is not None:
                if len(attr_value) == _field['oneof']:
                    attr_value = attr_value[_field['index']]
                # elif isinstance(attr_value, np.void):
                #    attr_value = attr_value[0]

        attrs[key] = (attr_value.decode('ascii')
                      if isinstance(attr_value, bytes) else attr_value)

    return attrs


def __get_coords(dset, data_sel: tuple) -> list:
    """
    Return coordinates of the HDF5 dataset with dimension scales

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the data is read
    data_sel :  tuple of slice(s)
       A numpy slice generated for example numpy.s_
       Default read while array

    Returns
    -------
    A sequence of tuples [(dims, data), ...]
    """
    coords = []
    if len(dset.dims) == dset.ndim:
        try:
            for ii, dim in enumerate(dset.dims):
                # get name of dimension
                name = PurePath(dim[0].name).name
                if name.startswith('row') or name.startswith('column'):
                    name = name.split(' ')[0]

                # determine coordinate
                buff = None
                if dim[0].size > 0 and not np.all(dim[0][()] == 0):
                    buff = dim[0][()]
                elif name in ('row', 'column'):
                    co_dtype = 'u2' if ((dset.shape[ii]-1) >> 16) == 0 else 'u4'
                    buff = np.arange(dset.shape[ii], dtype=co_dtype)

                if not (buff is None or data_sel is None):
                    buff = buff[data_sel[ii]]

                coords.append((name, buff))
        except RuntimeError:
            coords = []

    return coords


def __set_coords(dset, data_sel: tuple, dims: list) -> list:
    """
    Set coordinates of the HDF5 dataset

    Parameters
    ----------
    dset :  h5py.Dataset or numpy.array
       h5py dataset from which the data is read
    data_sel :  tuple of slice(s)
       A numpy slice generated for example numpy.s_
       Default read while array
    dims : list of strings
       Alternative names for the dataset dimensions if not attached to dataset
       Default coordinate names are ['time', ['row', ['column']]]

    Returns
    -------
    A sequence of tuples [(dims, data), ...]
    """
    if dims is None:
        if dset.ndim > 3:
            raise ValueError('not implemented for ndim > 3')

        dims = ['time', 'row', 'column'][-dset.ndim:]

    coords = []
    for ii in range(dset.ndim):
        co_dtype = 'u2' if ((dset.shape[ii]-1) >> 16) == 0 else 'u4'
        buff = np.arange(dset.shape[ii], dtype=co_dtype)
        if data_sel is not None:
            buff = buff[data_sel[ii]]
        coords.append((dims[ii], buff))

    return coords


def __get_data(dset, data_sel: tuple, field: str):
    """
    Return data of the HDF5 dataset

    Parameters
    ----------
    dset :  h5py.Dataset
       h5py dataset from which the data is read
    data_sel :  numpy slice
       A numpy slice generated for example numpy.s_
       Default read while array
    field : str
       Name of field in compound dataset or None

    Returns
    -------
    Numpy array

    Notes
    -----
    Read floats always as doubles
    """
    if data_sel is None:
        data_sel = ()

    if np.issubdtype(dset.dtype, np.floating):
        data = dset.astype(float)[data_sel]
        data[data == float.fromhex('0x1.ep+122')] = np.nan
        return data

    if field is None:
        return dset[data_sel]

    data = dset.fields(field)[data_sel]
    if np.issubdtype(data.dtype, np.floating):
        data = data.astype(float)
        data[data == float.fromhex('0x1.ep+122')] = np.nan
    return data


def __check_selection(data_sel: tuple, ndim: int) -> tuple:
    """
    Check and correct user provided data selection

    Notes
    -----
    If data_sel is used to select data from a dataset then the number of
    dimensions of data_sel should agree with the HDF5 dataset or one and
    only one Ellipsis has to be used.
    Thus allowed values for data_sel are:
    * [always]: (), np.s_[:], np.s_[...]
    * [1-D dataset]: np.s_[:-1], np.s_[0]
    * [2-D dataset]: np.s_[:-1, :], np.s_[0, :], np.s_[:-1, 0]
    * [3-D dataset]: np.s_[:-1, :, 2:4], np.s_[0, :, :], np.s_[:-1, 0, 2:4]
    * [Ellipsis] np.s_[0, ...], np.s_[..., 4], np.s_[0, ..., 4]
    """
    if data_sel in (np.s_[:], np.s_[...], np.s_[()]):
        return None

    if np.isscalar(data_sel):
        return (np.s_[data_sel:data_sel+1],)

    buff = ()
    for val in data_sel:
        if val == Ellipsis:
            for _ in range(ndim - len(data_sel) + 1):
                buff += np.index_exp[:]
        elif np.isscalar(val):
            buff += (np.s_[val:val+1],)
        else:
            buff += (val,)

    return buff


def h5_to_xr(h5_dset, data_sel=None, *, dims=None, field=None):
    """
    Create xarray::DataArray from a HDF5 dataset (with dimension scales)

    Parameters
    ----------
    h5_dset :  h5py.Dataset
       Data, dimensions, coordinates and attributes are read for this dataset
    data_sel :  slice, optional
       A numpy slice generated for example numpy.s_
    dims :  list of strings, optional
       Alternative names for the dataset dimensions if not attached to dataset
    field : str, optional
       Name of field in compound dataset or None

    Returns
    -------
    xarray.DataArray

    Examples
    --------
    >>> fid = h5py.File(flname, 'r')        # flname is a HDF5/netCDF4 file
    >>> xdata = h5_to_xr(fid['signal'])
    >>> fid.close()

    Combine Tropomi SWIR data of band7 and band8
    >>> fid = h5py.File(s5p_b7_prod, 'r')   # Tropomi band7 product
    >>> xdata7 = h5_to_xr(fid['signal'])
    >>> fid.close()
    >>> fid = h5py.File(s5p_b8_prod, 'r')   # Tropomi band8 product
    >>> xdata8 = h5_to_xr(fid['signal'])
    >>> fid.close()
    >>> xdata = xr.concat((xdata7, xdata8), dim='spectral_channel')

    Optionally, fix the 'column' dimension
    >>> xdata = xdata.assign_coords(
    >>> ... column=np.arange(xdata.column.size, dtype='u4'))
    """
    # Check data selection
    if data_sel is not None:
        data_sel = __check_selection(data_sel, h5_dset.ndim)

    # Name of this array
    name = PurePath(h5_dset.name).name

    # Values for this array
    data = __get_data(h5_dset, data_sel, field)

    # Coordinates (tick labels) to use for indexing along each dimension
    coords = []
    if dims is None:
        coords = __get_coords(h5_dset, data_sel)
    if not coords:
        coords = __set_coords(h5_dset, data_sel, dims)

    # - check if dimension of dataset and coordinates agree
    if data.ndim < len(coords):
        for ii in reversed(range(len(coords))):
            if np.isscalar(coords[ii][1]):
                del coords[ii]

    # - remove empty coordinates
    dims = []
    co_dict = {}
    for key, val in coords:
        dims.append(key)
        if val is not None:
            co_dict[key] = val

    # Attributes to assign to the array
    attrs = __get_attrs(h5_dset, field)

    return xr.DataArray(data,
                        coords=co_dict, dims=dims, name=name, attrs=attrs)


def data_to_xr(data, *, dims=None, name=None, long_name=None, units=None):
    """
    Create xarray::DataArray from a dataset

    Parameters
    ----------
    data :  array-like
       Data to be stored in the xarray
    dims :  list of strings
       Names for the dataset dimensions
    name : str
       A string that names the instance
    units :  str
       Units of the data, default: '1'
    long_name : str
       Long name describing the data, default: empty string

    Returns
    -------
    xarray.DataArray
    """
    coords = __set_coords(data, None, dims)
    attrs = {}
    attrs['units'] = '1' if units is None else units
    attrs['long_name'] = '' if long_name is None else long_name

    return xr.DataArray(data,
                        coords=coords, name=name, attrs=attrs)



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

This module

Functions
---------
fig_data_to_xarr(data, zscale=None, vperc=None, vrange=None, cmap=None)
   Prepare image data for plotting.

fig_qdata_to_xarr(data, ref_data=None, thres_worst=0.1, thres_bad=0.8,
                  qlabels=None)
   Prepare pixel-quality data for plotting.

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
from math import log10
import warnings

import numpy as np
import xarray as xr

import matplotlib.colors as mcolors
from matplotlib.ticker import AutoMinorLocator

from ..image_to_xarray import data_to_xr
from ..tol_colors import tol_cmap, tol_cset


# - local functions --------------------------------
def adjust_zunit(zunit: str, vmin: float, vmax: float):
    """
    Adjust units: electron to 'e' and Volt to 'V'
    and scale data range to <-1000, 1000>.

    Parameters
    ----------
    zunit :  str
       Units of the image data
    vmin, vmax : float
        image-data range

    Returns
    -------
    tuple:
        dscale, zunit
    """
    if zunit is None or zunit == '1':
        return 1, zunit

    if zunit.find('electron') >= 0:
        zunit = zunit.replace('electron', 'e')
    if zunit.find('Volt') >= 0:
        zunit = zunit.replace('Volt', 'V')
    if zunit.find('.s-1') >= 0:
        zunit = zunit.replace('.s-1', ' s$^{-1}$')

    if zunit[0] in ('e', 'V', 'A'):
        key_to_zunit = {-4: 'p', -3: 'n', -2: '\xb5', -1: 'm',
                        0: '', 1: 'k', 2: 'M', 3: 'G', 4: 'T'}
        max_value = max(abs(vmin), abs(vmax))
        key = min(4, max(-4, log10(max_value) // 3))

        return 1000 ** key, key_to_zunit[key] + zunit

    return 1, zunit


def set_norm(zscale: str, vmin: float, vmax: float):
    """
    Set data-range normalization
    """
    if zscale == 'log':
        return mcolors.LogNorm(vmin=max(vmin, 1e-6), vmax=vmax)

    if vmin == vmax:
        scale = max(1, abs(round((vmin + vmax) / 2)))
        vmin -= 1e-3 * scale
        vmax += 1e-3 * scale

    if zscale == 'diff':
        if vmin < 0 < vmax:
            vcntr = 0.
            tmp1, tmp2 = (vmin, vmax)
            vmin = -max(-tmp1, tmp2)
            vmax = max(-tmp1, tmp2)
        else:
            vcntr = (vmin + vmax) / 2
        return mcolors.TwoSlopeNorm(vcntr, vmin=vmin, vmax=vmax)

    if zscale == 'ratio':
        if vmin < 1 < vmax:
            vcntr = 1.
            tmp1, tmp2 = (vmin, vmax)
            vmin = min(tmp1, 1 / tmp2)
            vmax = max(1 / tmp1, tmp2)
        else:
            vcntr = (vmin + vmax) / 2
        return mcolors.TwoSlopeNorm(vcntr, vmin=vmin, vmax=vmax)

    return mcolors.Normalize(vmin=vmin, vmax=vmax)


def adjust_img_ticks(axx, xarr, dims=None):
    """
    Adjust ticks of the image axis
    """
    if dims is None or dims == 'X':
        if (xarr.shape[1] % 10) == 0:
            axx.set_xticks(np.linspace(0, xarr.shape[1], 6, dtype=int))
        elif (xarr.shape[1] % 8) == 0:
            axx.set_xticks(np.linspace(0, xarr.shape[1], 5, dtype=int))
        axx.xaxis.set_minor_locator(AutoMinorLocator())

    if dims is None or dims == 'Y':
        if (xarr.shape[0] % 10) == 0:
            axx.set_yticks(np.linspace(0, xarr.shape[0], 6, dtype=int))
        elif (xarr.shape[0] % 8) == 0:
            axx.set_yticks(np.linspace(0, xarr.shape[0], 5, dtype=int))
        axx.yaxis.set_minor_locator(AutoMinorLocator())


def fig_draw_panels(axx_p: dict, xarr, side_panels: str) -> None:
    """
    Draw two side-panels, one left and one under the main image panel

    Parameters
    ----------
    axx_p :  dict
       dictionary holding matplotlib Axes for panel 'X' and 'Y'
    xarr :  xarray.DataArray
       Object holding measurement data and attributes
    side_panels :  str
       Show row and column statistics in side plots.
    """
    cset = tol_cset('bright')

    # get numpy function to apply on image rows and columns for side pannels
    func_panels = {
        'median': np.median,
        'nanmedian': np.nanmedian,
        'mean': np.mean,
        'nanmean': np.nanmean,
        'quality': 'quality',
        'std': np.std,
        'nanstd': np.nanstd}.get(side_panels, None)
    if func_panels is None:
        raise KeyError(f'unknown function for side_panels: {side_panels}')

    # draw panel below the image pannel
    xdata = np.arange(xarr.shape[1])
    if side_panels == 'quality':
        ydata = np.sum(((xarr.values == 1) | (xarr.values == 2)), axis=0)
        axx_p['X'].step(xdata, ydata, linewidth=0.75, color=cset.yellow)
        ydata = np.sum((xarr.values == 1), axis=0)          # worst
        axx_p['X'].step(xdata, ydata, linewidth=0.75, color=cset.red)
        if len(xarr.attrs['flag_values']) == 6:
            ydata = np.sum((xarr.values == 4), axis=0)      # to_good
            axx_p['X'].step(xdata, ydata, linewidth=0.75, color=cset.green)
    else:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    r'All-NaN (slice|axis) encountered')
            axx_p['X'].plot(xdata, func_panels(xarr.values, axis=0),
                            linewidth=0.75, color=cset.blue)
    adjust_img_ticks(axx_p['X'], xarr, dims='X')
    axx_p['X'].grid()

    # draw panel left of the image pannel
    ydata = np.arange(xarr.shape[0])
    if side_panels == 'quality':
        xdata = np.sum(((xarr.values == 1) | (xarr.values == 2)), axis=1)
        axx_p['Y'].step(xdata, ydata, linewidth=0.75, color=cset.yellow)
        xdata = np.sum(xarr.values == 1, axis=1)            # worst
        axx_p['Y'].step(xdata, ydata, linewidth=0.75, color=cset.red)
        if len(xarr.attrs['flag_values']) == 6:
            xdata = np.sum(xarr.values == 4, axis=1)        # to_good
            axx_p['Y'].step(xdata, ydata, linewidth=0.75, color=cset.green)
    else:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    r'All-NaN (slice|axis) encountered')
            axx_p['Y'].plot(func_panels(xarr.values, axis=1), ydata,
                            linewidth=0.75, color=cset.blue)
    # axx_p['Y'].xaxis.tick_top()
    axx_p['X'].tick_params(axis='y', labelrotation=45, labelsize='small')
    axx_p['Y'].tick_params(axis='x', bottom=False, top=True,
                           labelbottom=False, labeltop=True,
                           labelrotation=-45, labelsize='small')
    adjust_img_ticks(axx_p['Y'], xarr, dims='Y')
    axx_p['Y'].grid()


# - main functions ---------------------------------
def fig_data_to_xarr(data, zscale=None, vperc=None, vrange=None):
    """
    Prepare image data for plotting.

    Parameters
    ----------
    data :  array-like or xarray.DataArray
    zscale :  str, default='linear'
        Scaling of the data values. Recognized values are: 'linear', 'log',
        'diff' or 'ratio'.
    vperc :  list, default=[1, 99]
        Range to normalize luminance data between percentiles min and max of
        array data.
    vrange :  list, default=None
        Range to normalize luminance data between vmin and vmax.

    Notes
    -----
    The input data should have two dimensions. If the input data is array-like,
    it will be converted into a xarray DataArray using the function data_to_xr.

    The returned xarray DataArray has addition attributes to facilitate the
    plotting methods of MONplot with names '_cmap' or start with '_z'.

    The default values of vperc are used when both vrange and vperc are None.
    When vrange and vperc are provided, then vrange is used.

    Returns
    -------
    xarray.DataArray
    """
    # make sure that we are working with a xarray DataArray
    xarr = data.copy() if isinstance(data, xr.DataArray) else data_to_xr(data)

    if zscale is None:
        zscale = 'linear'
    if zscale not in ('diff', 'linear', 'log', 'ratio'):
        raise RuntimeError(f'unknown zscale: {zscale}')
    xarr.attrs['_zscale'] = zscale

    # obtain image-data range
    if vrange is None and vperc is None:
        vmin, vmax = np.nanpercentile(xarr.values, (1., 99.))
    elif vrange is None:
        if len(vperc) != 2:
            raise TypeError('keyword vperc requires two values')
        vmin, vmax = np.nanpercentile(xarr.values, vperc)
    else:
        if len(vrange) != 2:
            raise TypeError('keyword vrange requires two values')
        vmin, vmax = vrange

    # set data units and scaling
    if 'units' in xarr.attrs:
        dscale, zunits = adjust_zunit(xarr.attrs['units'], vmin, vmax)
        xarr.values[np.isfinite(xarr.values)] /= dscale
        xarr.attrs['_zunits'] = zunits
    else:
        dscale = 1.
        xarr.attrs['_zunits'] = '1'

    # set data label
    if zscale == 'ratio' or xarr.attrs['_zunits'] == '1':
        xarr.attrs['_zlabel'] = {
            'ratio': 'ratio',
            'diff': 'difference'}.get(zscale, 'value')
    elif zscale == 'diff':
        xarr.attrs['_zlabel'] = f'difference [{xarr.attrs["_zunits"]}]'
    else:       # zscale in ('linear', 'log')
        xarr.attrs['_zlabel'] = f'value [{xarr.attrs["_zunits"]}]'

    # set matplotlib colormap
    xarr.attrs['_cmap'] = {'linear': tol_cmap('rainbow_PuRd'),
                           'log': tol_cmap('rainbow_WhBr'),
                           'diff': tol_cmap('sunset'),
                           'ratio': tol_cmap('sunset')}.get(zscale)

    # set matplotlib data normalization
    xarr.attrs['_znorm'] = set_norm(zscale, vmin / dscale, vmax / dscale)

    return xarr


# pylint: disable=too-many-arguments
def fig_qdata_to_xarr(data, ref_data=None, data_sel=None,
                      thres_worst=0.1, thres_bad=0.8, qlabels=None):
    """
    Prepare pixel-quality data for plotting

    Parameters
    ----------
    data :  array-like or xarray.DataArray
        Object holding detector pixel-quality data and attributes.
    ref_data :  numpy.ndarray, default=None
        Numpy array holding reference data, for example pixel quality
        reference map taken from the CKD. Shown are the changes with
        respect to the reference data.
    data_sel :  mask or index tuples for arrays, optional
        Select a region on the detector by fancy indexing (using a
        boolean/interger arrays), or using index tuples for arrays
        (generated with numpy.s_).
        Outside this region the pixels will be labeled: 'unusable'.
    thres_worst :  float, default=0.1
        Threshold to reject only the worst of the bad pixels, intended
        for CKD derivation.
    thres_bad :  float, default=0.8
        Threshold for bad pixels.
    qlabels : tuple of strings, optional
        Labels for the pixel-quality classes, see below

    Notes
    -----
    Without a reference dataset, the default quality ranking labels are:
    - 'unusable' :  pixels outside the illuminated region
    - 'worst'    :  0 <= value < thres_worst
    - 'bad'      :  0 <= value < thres_bad
    - 'good'     :  thres_bad <= value <= 1

    Otherwise the default quality ranking labels are:
    - 'unusable'    :  pixels outside the illuminated region
    - 'to worst'    :  from good or bad to worst
    - 'good to bad' :  from good to bad
    - 'to good'     :  from any rank to good
    - 'unchanged'   :  no change in rank

    Returns
    -------
    xarray.DataArray
    """
    if data_sel is None:
        exclude_region = None
    else:
        if isinstance(data_sel, np.ndarray):
            # pylint: disable=invalid-unary-operand-type
            exclude_region = ~data_sel
        else:
            exclude_region = np.full(data.shape, True)
            exclude_region[data_sel] = False

    def float_to_quality(arr):
        """
        Convert float value [0, 1] to quality classes
        """
        res = np.empty(arr.shape, dtype='i1')
        buff = arr.values if isinstance(arr, xr.DataArray) else arr
        res[buff >= thres_bad] = 4
        res[(buff > thres_worst) & (buff < thres_bad)] = 2
        res[buff <= thres_worst] = 1
        if exclude_region is not None:
            res[exclude_region] = 0
        return res

    qval = float_to_quality(data)
    if ref_data is not None:
        qdiff = float_to_quality(ref_data) - qval
        qval = np.full_like(qdiff, 8)
        qval[(qdiff == -2) | (qdiff == -3)] = 4
        qval[qdiff == 2] = 2
        qval[(qdiff == 1) | (qdiff == 3)] = 1
        if exclude_region is not None:
            qval[exclude_region] = 0

    # make sure that we are working with a xarray DataArray
    xarr = data_to_xr(qval)
    xarr.attrs['long_name'] = 'Pixel Quality'
    xarr.attrs['thres_bad'] = thres_bad
    xarr.attrs['thres_worst'] = thres_worst
    xarr.attrs['_zscale'] = 'quality'

    # define colors, data-range
    cset = tol_cset('bright')
    if ref_data is None:
        if qlabels is None:
            xarr.attrs['flag_meanings'] = ("unusable", "worst", "bad", "good")
        elif len(qlabels) != 4:
            raise TypeError('keyword qlabels requires four labels')
        else:
            xarr.attrs['flag_meanings'] = qlabels
        # define colors for resp. unusable, worst, bad and good
        ctuple = (cset.grey, cset.red, cset.yellow, '#FFFFFF')
        xarr.attrs['valid_range'] = np.array([0, 8], dtype='i1')
        xarr.attrs['flag_values'] = np.array([0, 1, 2, 4, 8], dtype='i1')
    else:
        if qlabels is None:
            xarr.attrs['flag_meanings'] = ("unusable", "to worst",
                                           "good to bad ",
                                           "to good", "unchanged")
        elif len(qlabels) != 5:
            raise TypeError('keyword qlabels requires five labels')
        else:
            xarr.attrs['flag_meanings'] = qlabels
        # define colors for resp. unusable, worst, bad, good and unchanged
        ctuple = (cset.grey, cset.red, cset.yellow, cset.green, '#FFFFFF')
        xarr.attrs['valid_range'] = np.array([0, 16], dtype='i1')
        xarr.attrs['flag_values'] = np.array([0, 1, 2, 4, 8, 16], dtype='i1')

    xarr.attrs['_cmap'] = mcolors.ListedColormap(ctuple)
    xarr.attrs['_znorm'] = mcolors.BoundaryNorm(xarr.attrs['flag_values'],
                                                xarr.attrs['_cmap'].N)
    return xarr



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

Definition of colour schemes for lines and maps that also work for colour-blind
people. See https://personal.sron.nl/~pault/ for background information and
best usage of the schemes.

Reference
---------
   https://personal.sron.nl/~pault/

Copyright (c) 2019-2022 Paul Tol (SRON)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
from typing import NamedTuple

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.colors import LinearSegmentedColormap, to_rgba_array


def discretemap(colormap, hexclrs):
    """
    Produce a colormap from a list of discrete colors without interpolation.
    """
    clrs = to_rgba_array(hexclrs)
    clrs = np.vstack([clrs[0], clrs, clrs[-1]])
    cdict = {}
    for ii, key in enumerate(('red', 'green', 'blue')):
        cdict[key] = [(jj / (len(clrs)-2.), clrs[jj, ii], clrs[jj+1, ii])
                      for jj in range(len(clrs)-1)]
    return LinearSegmentedColormap(colormap, cdict)


# pylint: disable=invalid-name
class TOLcmaps():
    """
    Class TOLcmaps definition.
    """
    def __init__(self):
        """
        """
        self.cmap = None
        self.cname = None
        self.namelist = (
            'sunset_discrete', 'sunset', 'BuRd_discrete', 'BuRd',
            'PRGn_discrete', 'PRGn', 'YlOrBr_discrete', 'YlOrBr', 'WhOrBr',
            'iridescent', 'incandescent', 'rainbow_PuRd', 'rainbow_PuBr',
            'rainbow_WhRd', 'rainbow_WhBr', 'rainbow_WhBr_condense',
            'rainbow_discrete')

        self.funcdict = dict(
            zip(self.namelist,
                (self.__sunset_discrete, self.__sunset, self.__BuRd_discrete,
                 self.__BuRd, self.__PRGn_discrete, self.__PRGn,
                 self.__YlOrBr_discrete, self.__YlOrBr, self.__WhOrBr,
                 self.__iridescent, self.__incandescent,
                 self.__rainbow_PuRd, self.__rainbow_PuBr,
                 self.__rainbow_WhRd, self.__rainbow_WhBr,
                 self.__rainbow_WhBr_condense, self.__rainbow_discrete)))

    def __sunset_discrete(self):
        """
        Define colormap 'sunset_discrete'.
        """
        clrs = ['#364B9A', '#4A7BB7', '#6EA6CD', '#98CAE1', '#C2E4EF',
                '#EAECCC', '#FEDA8B', '#FDB366', '#F67E4B', '#DD3D2D',
                '#A50026']
        self.cmap = discretemap(self.cname, clrs)
        self.cmap.set_bad('#FFFFFF')

    def __sunset(self):
        """
        Define colormap 'sunset'.
        """
        clrs = ['#364B9A', '#4A7BB7', '#6EA6CD', '#98CAE1', '#C2E4EF',
                '#EAECCC', '#FEDA8B', '#FDB366', '#F67E4B', '#DD3D2D',
                '#A50026']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#FFFFFF')

    def __BuRd_discrete(self):
        """
        Define colormap 'BuRd_discrete'.
        """
        clrs = ['#2166AC', '#4393C3', '#92C5DE', '#D1E5F0', '#F7F7F7',
                '#FDDBC7', '#F4A582', '#D6604D', '#B2182B']
        self.cmap = discretemap(self.cname, clrs)
        self.cmap.set_bad('#FFEE99')

    def __BuRd(self):
        """
        Define colormap 'BuRd'.
        """
        clrs = ['#2166AC', '#4393C3', '#92C5DE', '#D1E5F0', '#F7F7F7',
                '#FDDBC7', '#F4A582', '#D6604D', '#B2182B']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#FFEE99')

    def __PRGn_discrete(self):
        """
        Define colormap 'PRGn_discrete'.
        """
        clrs = ['#762A83', '#9970AB', '#C2A5CF', '#E7D4E8', '#F7F7F7',
                '#D9F0D3', '#ACD39E', '#5AAE61', '#1B7837']
        self.cmap = discretemap(self.cname, clrs)
        self.cmap.set_bad('#FFEE99')

    def __PRGn(self):
        """
        Define colormap 'PRGn'.
        """
        clrs = ['#762A83', '#9970AB', '#C2A5CF', '#E7D4E8', '#F7F7F7',
                '#D9F0D3', '#ACD39E', '#5AAE61', '#1B7837']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#FFEE99')

    def __YlOrBr_discrete(self):
        """
        Define colormap 'YlOrBr_discrete'.
        """
        clrs = ['#FFFFE5', '#FFF7BC', '#FEE391', '#FEC44F', '#FB9A29',
                '#EC7014', '#CC4C02', '#993404', '#662506']
        self.cmap = discretemap(self.cname, clrs)
        self.cmap.set_bad('#888888')

    def __YlOrBr(self):
        """
        Define colormap 'YlOrBr'.
        """
        clrs = ['#FFFFE5', '#FFF7BC', '#FEE391', '#FEC44F', '#FB9A29',
                '#EC7014', '#CC4C02', '#993404', '#662506']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#888888')

    def __WhOrBr(self):
        """
        Define colormap 'WhOrBr'.
        """
        clrs = ['#FFFFFF', '#FFF7BC', '#FEE391', '#FEC44F', '#FB9A29',
                '#EC7014', '#CC4C02', '#993404', '#662506']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#888888')

    def __iridescent(self):
        """
        Define colormap 'Iridescent'.
        """
        clrs = ['#FEFBE9', '#FCF7D5', '#F5F3C1', '#EAF0B5', '#DDECBF',
                '#D0E7CA', '#C2E3D2', '#B5DDD8', '#A8D8DC', '#9BD2E1',
                '#8DCBE4', '#81C4E7', '#7BBCE7', '#7EB2E4', '#88A5DD',
                '#9398D2', '#9B8AC4', '#9D7DB2', '#9A709E', '#906388',
                '#805770', '#684957', '#46353A']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#999999')

    def __incandescent(self):
        """
        Define colormap 'Incandescent'.
        """
        clrs = ['#CEFFFF', '#C6F7D6', '#A2F49B', '#BBE453', '#D5CE04',
                '#E7B503', '#F19903', '#F6790B', '#F94902', '#E40515',
                '#A80003']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#888888')

    def __rainbow_PuRd(self):
        """
        Define colormap 'rainbow_PuRd'.
        """
        clrs = ['#6F4C9B', '#6059A9', '#5568B8', '#4E79C5', '#4D8AC6',
                '#4E96BC', '#549EB3', '#59A5A9', '#60AB9E', '#69B190',
                '#77B77D', '#8CBC68', '#A6BE54', '#BEBC48', '#D1B541',
                '#DDAA3C', '#E49C39', '#E78C35', '#E67932', '#E4632D',
                '#DF4828', '#DA2222']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#FFFFFF')

    def __rainbow_PuBr(self):
        """
        Define colormap 'rainbow_PuBr'.
        """
        clrs = ['#6F4C9B', '#6059A9', '#5568B8', '#4E79C5', '#4D8AC6',
                '#4E96BC', '#549EB3', '#59A5A9', '#60AB9E', '#69B190',
                '#77B77D', '#8CBC68', '#A6BE54', '#BEBC48', '#D1B541',
                '#DDAA3C', '#E49C39', '#E78C35', '#E67932', '#E4632D',
                '#DF4828', '#DA2222', '#B8221E', '#95211B', '#721E17',
                '#521A13']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#FFFFFF')

    def __rainbow_WhRd(self):
        """
        Define colormap 'rainbow_WhRd'.
        """
        clrs = ['#E8ECFB', '#DDD8EF', '#D1C1E1', '#C3A8D1', '#B58FC2',
                '#A778B4', '#9B62A7', '#8C4E99', '#6F4C9B', '#6059A9',
                '#5568B8', '#4E79C5', '#4D8AC6', '#4E96BC', '#549EB3',
                '#59A5A9', '#60AB9E', '#69B190', '#77B77D', '#8CBC68',
                '#A6BE54', '#BEBC48', '#D1B541', '#DDAA3C', '#E49C39',
                '#E78C35', '#E67932', '#E4632D', '#DF4828', '#DA2222']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#666666')

    def __rainbow_WhBr(self):
        """
        Define colormap 'rainbow_WhBr'.
        """
        clrs = ['#E8ECFB', '#DDD8EF', '#D1C1E1', '#C3A8D1', '#B58FC2',
                '#A778B4', '#9B62A7', '#8C4E99', '#6F4C9B', '#6059A9',
                '#5568B8', '#4E79C5', '#4D8AC6', '#4E96BC', '#549EB3',
                '#59A5A9', '#60AB9E', '#69B190', '#77B77D', '#8CBC68',
                '#A6BE54', '#BEBC48', '#D1B541', '#DDAA3C', '#E49C39',
                '#E78C35', '#E67932', '#E4632D', '#DF4828', '#DA2222',
                '#B8221E', '#95211B', '#721E17', '#521A13']
        self.cmap = LinearSegmentedColormap.from_list(self.cname, clrs)
        self.cmap.set_bad('#666666')

    def __rainbow_WhBr_condense(self):
        """
        Define colormap 'rainbow_WhBr_condense'.
        """
        clrs = ['#E8ECFB', '#E4E5F7', '#E1DFF3', '#DDD8EF', '#D9D0EA',
                '#D5C8E5', '#D1C1E1', '#CCB8DB', '#C7B0D6', '#C3A8D1',
                '#BEA0CC', '#BA97C7', '#B58FC2', '#B087BD', '#AC80B8',
                '#A778B4', '#A371AF', '#9F6AAB', '#9B62A7', '#965CA2',
                '#91559E', '#8C4E99', '#824D9A', '#794C9A', '#6F4C9B',
                '#6A509F', '#6555A4', '#6059A9', '#5C5EAE', '#5963B3',
                '#5568B8', '#526EBC', '#5073C0', '#4E79C5', '#4E7FC5',
                '#4D84C6', '#4D8AC6', '#4D8EC3', '#4D92C0', '#4E96BC',
                '#5099B9', '#529CB6', '#549EB3', '#56A1AF', '#58A3AC',
                '#59A5A9', '#5CA7A5', '#5EA9A1', '#60AB9E', '#63AD99',
                '#66AF94', '#69B190', '#6EB38A', '#73B584', '#77B77D',
                '#7EB976', '#85BA6F', '#8CBC68', '#95BD61', '#9DBD5B',
                '#A6BE54', '#AEBD50', '#B6BD4C', '#BEBC48', '#C4BA45',
                '#CAB843', '#D1B541', '#D5B23F', '#D9AE3E', '#DDAA3C',
                '#DFA63B', '#E1A13A', '#E49C39', '#E59738', '#E69137',
                '#E78C35', '#E68534', '#E67F33', '#E67932', '#E57130',
                '#E56A2F', '#E4632D', '#E25A2C', '#E1512A', '#DF4828',
                '#DE3B26', '#DC2F24', '#DA2222', '#CF2221', '#C42220',
                '#B8221E', '#AC221D', '#A1211C', '#95211B', '#892019',
                '#7E1F18', '#721E17', '#681C15', '#5D1B14', '#521A13']
        self.cmap = discretemap(self.cname, clrs)
        self.cmap.set_bad('#666666')

    def __rainbow_discrete(self, lut=None):
        """
        Define colormap 'rainbow_discrete'.
        """
        clrs = ['#E8ECFB', '#D9CCE3', '#D1BBD7', '#CAACCB', '#BA8DB4',
                '#AE76A3', '#AA6F9E', '#994F88', '#882E72', '#1965B0',
                '#437DBF', '#5289C7', '#6195CF', '#7BAFDE', '#4EB265',
                '#90C987', '#CAE0AB', '#F7F056', '#F7CB45', '#F6C141',
                '#F4A736', '#F1932D', '#EE8026', '#E8601C', '#E65518',
                '#DC050C', '#A5170E', '#72190E', '#42150A']
        indexes = [[9], [9, 25], [9, 17, 25], [9, 14, 17, 25],
                   [9, 13, 14, 17, 25], [9, 13, 14, 16, 17, 25],
                   [8, 9, 13, 14, 16, 17, 25], [8, 9, 13, 14, 16, 17, 22, 25],
                   [8, 9, 13, 14, 16, 17, 22, 25, 27],
                   [8, 9, 13, 14, 16, 17, 20, 23, 25, 27],
                   [8, 9, 11, 13, 14, 16, 17, 20, 23, 25, 27],
                   [2, 5, 8, 9, 11, 13, 14, 16, 17, 20, 23, 25],
                   [2, 5, 8, 9, 11, 13, 14, 15, 16, 17, 20, 23, 25],
                   [2, 5, 8, 9, 11, 13, 14, 15, 16, 17, 19, 21, 23, 25],
                   [2, 5, 8, 9, 11, 13, 14, 15, 16, 17, 19, 21, 23, 25, 27],
                   [2, 4, 6, 8, 9, 11, 13, 14, 15, 16, 17, 19, 21, 23, 25, 27],
                   [2, 4, 6, 7, 8, 9, 11, 13, 14, 15, 16, 17, 19, 21, 23, 25,
                    27],
                   [2, 4, 6, 7, 8, 9, 11, 13, 14, 15, 16, 17, 19, 21, 23, 25,
                    26, 27],
                   [1, 3, 4, 6, 7, 8, 9, 11, 13, 14, 15, 16, 17, 19, 21, 23,
                    25, 26, 27],
                   [1, 3, 4, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 19, 21,
                    23, 25, 26, 27],
                   [1, 3, 4, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 20,
                    22, 24, 25, 26, 27],
                   [1, 3, 4, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 20,
                    22, 24, 25, 26, 27, 28],
                   [0, 1, 3, 4, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18,
                    20, 22, 24, 25, 26, 27, 28]]
        if lut is None or lut < 1 or lut > 23:
            lut = 22
        self.cmap = discretemap(self.cname, [clrs[i] for i in indexes[lut-1]])
        if lut == 23:
            self.cmap.set_bad('#777777')
        else:
            self.cmap.set_bad('#FFFFFF')

    def show(self):
        """
        List names of defined colormaps.
        """
        print(' '.join(repr(n) for n in self.namelist))

    def get(self, cname='rainbow_PuRd', lut=None):
        """
        Return requested colormap, default is 'rainbow_PuRd'.
        """
        self.cname = cname
        if cname == 'rainbow_discrete':
            self.__rainbow_discrete(lut)
        else:
            self.funcdict[cname]()
        return self.cmap


def tol_cmap(colormap=None, lut=None):
    """
    Continuous and discrete color sets for ordered data.

    Return a matplotlib colormap.
    Parameter lut is ignored for all colormaps except 'rainbow_discrete'.
    """
    obj = TOLcmaps()
    if colormap is None:
        return obj.namelist
    if colormap not in obj.namelist:
        colormap = 'rainbow_PuRd'
        print('*** Warning: requested colormap not defined,',
              f'known colormaps are {obj.namelist}.',
              f'Using {colormap}.')
    return obj.get(colormap, lut)


def tol_cset(colorset=None):
    """
    Discrete color sets for qualitative data.

    Define a namedtuple instance with the colors.
    Examples for: cset = tol_cset(<scheme>)
      - cset.red and cset[1] give the same color (in default 'bright' colorset)
      - cset._fields gives a tuple with all color names
      - list(cset) gives a list with all colors
    """
    namelist = ('bright', 'high-contrast', 'vibrant', 'muted', 'light')
    if colorset is None:
        return namelist

    if colorset not in namelist:
        print('*** Warning: requested colorset not defined,',
              f'known colorsets are {namelist}.',
              f'Using {colorset}.')

    if colorset == 'high-contrast':
        cset = NamedTuple('Hcset', [('blue', str), ('yellow', str),
                                    ('red', str), ('black', str)])
        return cset('#004488', '#DDAA33', '#BB5566', '#000000')

    if colorset == 'vibrant':
        cset = NamedTuple('Vcset', [('orange', str), ('blue', str),
                                    ('cyan', str), ('magenta', str),
                                    ('red', str), ('teal', str),
                                    ('grey', str), ('black', str)])
        return cset('#EE7733', '#0077BB', '#33BBEE', '#EE3377', '#CC3311',
                    '#009988', '#BBBBBB', '#000000')

    if colorset == 'muted':
        cset = NamedTuple('Mcset', [('rose', str), ('indigo', str),
                                    ('sand', str), ('green', str),
                                    ('cyan', str), ('wine', str),
                                    ('teal', str), ('olive', str),
                                    ('purple', str), ('pale_grey', str),
                                    ('black', str)])
        return cset('#CC6677', '#332288', '#DDCC77', '#117733', '#88CCEE',
                    '#882255', '#44AA99', '#999933', '#AA4499', '#DDDDDD',
                    '#000000')

    if colorset == 'light':
        cset = NamedTuple('Lcset', [('light_blue', str), ('orange', str),
                                    ('light_yellow', str), ('pink', str),
                                    ('light_cyan', str), ('mint', str),
                                    ('pear', str), ('olive', str),
                                    ('pale_grey', str), ('black', str)])
        return cset('#77AADD', '#EE8866', '#EEDD88', '#FFAABB', '#99DDFF',
                    '#44BB99', '#BBCC33', '#AAAA00', '#DDDDDD', '#000000')

    # Default: return colorset is 'bright'
    cset = NamedTuple('Bcset', [('blue', str), ('red', str),
                                ('green', str), ('yellow', str),
                                ('cyan', str), ('purple', str),
                                ('grey', str), ('black', str)])
    return cset('#4477AA', '#EE6677', '#228833', '#CCBB44', '#66CCEE',
                '#AA3377', '#BBBBBB', '#000000')


def main():
    """
    Show all available colormaps and colorsets
    """
    # Change default colorset (for lines) and colormap (for maps).
#    plt.rc('axes', prop_cycle=plt.cycler('color', list(tol_cset('bright'))))
#    plt.cm.register_cmap('rainbow_PuRd', tol_cmap('rainbow_PuRd'))
#    plt.rc('image', cmap='rainbow_PuRd')

    # Show colorsets tol_cset(<scheme>).
    schemes = tol_cset()
    fig, axes = plt.subplots(ncols=len(schemes))
    fig.subplots_adjust(top=0.92, bottom=0.02, left=0.0, right=0.91)
    for ax, scheme in zip(axes, schemes):
        cset = tol_cset(scheme)
        names = cset._fields
        colors = list(cset)
        for name, color in zip(names, colors):
            ax.scatter([], [], c=color, s=80, label=name)
        ax.set_axis_off()
        ax.legend(loc=2)
        ax.set_title(scheme)
    plt.show()

    # Show colormaps tol_cmap(<scheme>).
    schemes = tol_cmap()
    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient, gradient))
    fig, axes = plt.subplots(nrows=len(schemes))
    fig.subplots_adjust(top=0.98, bottom=0.02, left=0.29, right=0.99)
    for ax, scheme in zip(axes, schemes):
        pos = list(ax.get_position().bounds)
        ax.set_axis_off()
        ax.imshow(gradient, aspect=4, cmap=tol_cmap(scheme))
        fig.text(pos[0] - 0.01, pos[1] + pos[3]/2.,
                 scheme, va='center', ha='right', fontsize=10)
    plt.show()

    # Show colormaps tol_cmap('rainbow_discrete', <lut>).
    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient, gradient))
    fig, axes = plt.subplots(nrows=23)
    fig.subplots_adjust(top=0.98, bottom=0.02, left=0.25, right=0.99)
    for lut, ax in enumerate(axes, start=1):
        pos = list(ax.get_position().bounds)
        ax.set_axis_off()
        ax.imshow(gradient, aspect=4, cmap=tol_cmap('rainbow_discrete', lut))
        fig.text(pos[0] - 0.01, pos[1] + pos[3]/2.,
                 'rainbow_discrete, ' + str(lut),
                 va='center', ha='right', fontsize=10)
    plt.show()


if __name__ == '__main__':
    main()



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

Performs unit-tests on MONplot methods: draw_signal, draw_quality,
   draw_trend and draw_lplot ## draw_cmp_images

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
from datetime import datetime, timedelta

import numpy as np
import xarray as xr

from moniplot.image_to_xarray import data_to_xr
from moniplot.lib.fig_info import FIGinfo
from moniplot.mon_plot import MONplot


def get_test_data(data_sel=None, xy_min=-5, xy_max=5, delta=0.01, error=0):
    """
    Generate synthetic data to simulate a square-detector image
    """
    if data_sel is None:
        data_sel = [(), ()]

    res = np.arange(xy_min, xy_max, delta)
    xmesh, ymesh = np.meshgrid(res[data_sel[1]], res[data_sel[0]])
    zz1 = np.exp(-xmesh ** 2 - ymesh ** 2)
    zz2 = np.exp(-(xmesh - 1) ** 2 - (ymesh - 1) ** 2)
    data = (zz1 - zz2) * 2
    data += np.random.default_rng().normal(0., error, data.shape)

    return data_to_xr(data, long_name='bogus data', units='Volt')


def run_draw_signal(plot):
    """
    Run unit tests on MONplot::draw_signal
    """
    print('Run unit tests on MONplot::draw_signal')
    msm = get_test_data(error=.1)
    # msm_ref = get_test_data(error=0.025)

    # image aspect=4
    plot.draw_signal(msm, zscale='linear', side_panels='none',
                     title='method=linear; aspect=1; fig_pos=above')
    plot.draw_signal(msm, zscale='linear',
                     title='method=linear; aspect=1; fig_pos=above')
    plot.draw_signal(msm, zscale='diff',
                     title='method=diff; aspect=1; fig_pos=above')
    msm1 = msm.copy()
    msm1.values = (msm1.values + 3) / 3
    msm1.attrs['units'] = '1'
    plot.draw_signal(msm1, zscale='ratio',
                     title='method=ratio; aspect=1; fig_pos=above')
    plot.draw_signal(np.abs(msm), zscale='log', side_panels='none',
                     title='method=error; aspect=1; fig_pos=above')
    # plot.draw_signal(msm, fig_info=fig_info_in.copy(),
    #                 title='method=linear; aspect=1; fig_pos=right')

    # image aspect=2
    msm = get_test_data(data_sel=np.s_[500-250:500+250, :], error=.1)
    plot.draw_signal(msm, side_panels='none',
                     title='method=linear; aspect=2; fig_pos=above')
    plot.draw_signal(msm,
                     title='method=linear; aspect=2; fig_pos=above')
    # plot.draw_signal(msm, fig_info=fig_info_in.copy(),
    #                 title='method=linear; aspect=2; fig_pos=right')

    # image aspect=3
    msm = get_test_data(data_sel=np.s_[500-125:500+125, 500-375:500+375],
                        error=.1)
    plot.draw_signal(msm, side_panels='none',
                     title='method=linear; aspect=3; fig_pos=above')
    plot.draw_signal(msm,
                     title='method=linear; aspect=3; fig_pos=above')
    # plot.draw_signal(msm,
    #                 fig_info=fig_info_in.copy(),
    #                 title='method=linear; aspect=3; fig_pos=right')

    # image aspect=4
    msm = get_test_data(data_sel=np.s_[500-128:500+128, :], error=.1)
    plot.draw_signal(msm, side_panels='none',
                     title='method=linear; aspect=4; fig_pos=above')
    plot.draw_signal(msm,
                     title='method=linear; aspect=4; fig_pos=above')
    # plot.draw_signal(msm, fig_info=fig_info_in.copy(),
    #                 title='method=linear; aspect=4; fig_pos=right')


def run_draw_quality(plot):
    """
    Run unit tests on MONplot::draw_quality
    """
    print('Run unit tests on MONplot::draw_quality')
    row = np.linspace(0, 1., 1000)
    ref_data = np.repeat(row[None, :], 256, axis=0)

    data = ref_data.copy()
    data[125:175, 30:50] = 0.4
    data[125:175, 50:70] = 0.95
    data[75:125, 250:300] = 0.05
    data[125:175, 600:650] = 0.95
    data[75:125, 900:925] = 0.05
    data[75:125, 825:875] = 0.4

    figinfo_in = FIGinfo()
    figinfo_in.add('orbits', (17, [23662, 23707]), fmt='{} in {}')
    figinfo_in.add('coverage', ('2022-05-08', '2022-05-10'), fmt='{} / {}')

    plot.draw_quality(data, data_sel=np.s_[11:228, 16:991],
                      title='no reference', fig_info=figinfo_in.copy())
    plot.draw_quality(data, ref_data=ref_data, data_sel=np.s_[11:228, 16:991],
                      title='with reference', fig_info=figinfo_in.copy())
    plot.draw_quality(data, data_sel=np.s_[11:228, 16:991],
                      side_panels='none', title='no reference',
                      fig_info=figinfo_in.copy())


def run_draw_trend(plot):
    """
    Run unit tests on MONplot::draw_trend
    """
    print('Run unit tests on MONplot::draw_trend')
    n_elmnt = 200
    xval = np.arange(n_elmnt) / 100

    res = []
    hk_dtype = np.dtype([('mean', 'f8'), ('err1', 'f8'), ('err2', 'f8')])
    buff = np.empty(len(xval), dtype=hk_dtype)
    data = 140. + (100 - np.arange(n_elmnt)) / 1000
    buff['mean'] = data
    buff['err1'] = data - .0125
    buff['err2'] = data + .0075
    hk_attrs = {'long_name': 'SWIR detector temperature', 'units': 'K'}
    res.append(xr.DataArray(buff, name='detector_temp', attrs=hk_attrs,
                            coords={'orbit': np.arange(n_elmnt)}))

    buff = np.empty(len(xval), dtype=hk_dtype)
    data = 202.1 + (100 - np.arange(n_elmnt)) / 1000
    buff['mean'] = data
    buff['err1'] = data - .15
    buff['err2'] = data + .175
    hk_attrs = {'long_name': 'SWIR grating temperature', 'units': 'K'}
    res.append(xr.DataArray(buff, name='grating_temp', attrs=hk_attrs,
                            coords={'orbit': np.arange(n_elmnt)}))

    buff = np.empty(len(xval), dtype=hk_dtype)
    data = 208.2 + (100 - np.arange(n_elmnt)) / 1000
    buff['mean'] = data
    buff['err1'] = data - .15
    buff['err2'] = data + .175
    hk_attrs = {'long_name': 'SWIR OBM temperature', 'units': 'K'}
    res.append(xr.DataArray(buff, name='obm_temp', attrs=hk_attrs,
                            coords={'orbit': np.arange(n_elmnt)}))
    hk_ds = xr.merge(res, combine_attrs="drop_conflicts")

    msm1 = data_to_xr(np.sin(xval * np.pi), name='msm1', dims=['orbit'])
    msm2 = data_to_xr(np.cos(xval * np.pi), name='msm2', dims=['orbit'])
    msm_ds = xr.merge((msm1, msm2), combine_attrs="drop_conflicts")

    # plot.draw_trend(msm_ds, title='one dataset, no house-keeping')
    plot.draw_trend(xds=msm_ds, title='two datasets, no house-keeping')
    plot.draw_trend(hk_xds=hk_ds, title='no datasets, only house-keeping')
    plot.draw_trend(msm_ds, hk_ds, title='two datasets and house-keeping')


def run_draw_lplot(plot):
    """
    Run unit tests on MONplot::draw_lplot
    """
    print('Run unit tests on MONplot::draw_lplot')
    xval = np.arange(200) / 100
    plot.draw_lplot(xval, np.sin(xval * np.pi), color=0,
                    label='sinus', marker='o', linestyle='-')
    plot.draw_lplot(xval, np.cos(xval * np.pi), color=1,
                    label='cosinus', marker='o', linestyle='-')
    plot.draw_lplot(None, None, ylim=[-1.05, 1.05],
                    xlabel='x-axis [Pi]', ylabel='y-axis',
                    title='draw_lplot [no time_axis]')

    xval = np.arange(500) / 100
    plot.draw_lplot(xval, np.sin(xval * np.pi), color=0,
                    label='sinus', marker='o', linestyle='-')
    plot.draw_lplot(xval, np.cos(xval * np.pi), color=1,
                    label='cosinus', marker='o', linestyle='-')
    plot.draw_lplot(None, None, ylim=[-1.05, 1.05],
                    xlabel='x-axis [Pi]', ylabel='y-axis',
                    title='draw_lplot [no time_axis]')

    customdate = datetime(2016, 1, 1, 13, 0, 0)
    yval = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20]
    xval = [customdate + timedelta(hours=i, minutes=4*i)
            for i in range(len(yval))]
    plot.draw_lplot(xval, yval, color=0, label='mydata',
                    marker='o', linestyle='-')
    plot.draw_lplot(None, None, title='draw_lplot [time_axis]',
                    xlabel='x-axis', ylabel='y-axis')


def run_draw_qhist(plot):
    """
    Run unit tests on MONplot::draw_qhist
    """
    print('Run unit tests on MONplot::draw_qhist')
    buff0 = np.repeat(0.9 + np.random.rand(1000) / 10, 56)
    buff1 = np.repeat(np.random.rand(1000) / 10, 10)
    buff2 = 0.1 + np.random.rand(1000) / 10
    buff3 = np.repeat(0.2 + np.random.rand(1000) / 10, 2)
    buff4 = np.repeat(0.3 + np.random.rand(1000) / 10, 3)
    buff5 = np.repeat(0.4 + np.random.rand(1000) / 10, 4)
    buff6 = np.repeat(0.5 + np.random.rand(1000) / 10, 8)
    buff7 = np.repeat(0.6 + np.random.rand(1000) / 10, 12)
    buff8 = np.repeat(0.7 + np.random.rand(1000) / 10, 20)
    buff9 = np.repeat(0.8 + np.random.rand(1000) / 10, 40)
    buffa = np.repeat(0.9 + np.random.rand(1000) / 10, 100)
    frame = np.concatenate((buff0, buff1, buff2, buff3, buff4,
                            buff5, buff6, buff7, buff8, buff9,
                            buffa)).reshape(256, 1000)
    msm = xr.merge([data_to_xr(frame, name='dpqm',
                               long_name='pixel-quality map'),
                    data_to_xr(frame, name='dpqm_dark',
                               long_name='pixel-quality map (dark)'),
                    data_to_xr(frame, name='dpqm_noise',
                               long_name='pixel-quality map (noise average)'),
                    data_to_xr(frame, name='dpqm_noise_var',
                               long_name='pixel-quality map (noise variance)')])

    plot.draw_qhist(msm, data_sel=np.s_[11:228, 16:991], title='my histogram')


# --------------------------------------------------
def main():
    """
    main function
    """
    check_draw_signal = True
    check_draw_quality = True
    check_draw_qhist = True
    check_draw_trend = True
    check_draw_lplot = True

    # ---------- UNIT TEST: draw_signal ----------
    if check_draw_signal:
        plot = MONplot('mon_plot_draw_signal.pdf',
                       caption='Unit test of MONplot [draw_signal]')
        plot.set_institute('SRON')
        run_draw_signal(plot)
        plot.close()

    # ---------- UNIT TEST: draw_quality ----------
    if check_draw_quality:
        plot = MONplot('mon_plot_draw_quality.pdf',
                       caption='Unit test of MONplot [draw_quality]')
        plot.set_institute('SRON')
        run_draw_quality(plot)
        plot.close()

    # ---------- UNIT TEST: draw_qhist ----------
    if check_draw_qhist:
        plot = MONplot('mon_plot_draw_qhist.pdf',
                       caption='Unit test of MONplot [draw_qhist]')
        plot.set_institute('SRON')
        run_draw_qhist(plot)
        plot.close()

    # ---------- UNIT TEST: draw_trend ----------
    if check_draw_trend:
        plot = MONplot('mon_plot_draw_trend.pdf',
                       caption='Unit test of MONplot [draw_trend]')
        plot.set_institute('SRON')
        run_draw_trend(plot)
        plot.close()

    # ---------- UNIT TEST: draw_lplot ----------
    if check_draw_lplot:
        plot = MONplot('mon_plot_draw_lplot.pdf',
                       caption='Unit test of MONplot [draw_lplot]')
        plot.set_institute('SRON')
        run_draw_lplot(plot)
        plot.close()


# - main code --------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

The class MONplot contains generic plot functions.

Notes
-----
The methods of the class MONplot will accept numpy arrays as input and display
your data without knowledge on the data units and coordinates. In most cases,
this will be enough for a quick inspection of your data.
However, when you use xarray labeled arrays and datasets then the software
will use the name of the xarray arrays, coordinate names and data attributes,
such as 'long_name' and 'units'.

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
from datetime import datetime
from pathlib import PurePath

import numpy as np
import xarray as xr

try:
    from cartopy import crs as ccrs
except ModuleNotFoundError:
    FOUND_CARTOPY = False
else:
    FOUND_CARTOPY = True
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.gridspec import GridSpec

from .biweight import biweight

from .lib.fig_info import FIGinfo
from .lib.fig_draw_image import (adjust_img_ticks,
                                 fig_data_to_xarr,
                                 fig_qdata_to_xarr,
                                 fig_draw_panels)
from .lib.fig_draw_trend import add_subplot, add_hk_subplot
from .lib.fig_draw_qhist import fig_draw_qhist
from .lib.fig_draw_lplot import fig_draw_lplot, close_draw_lplot
from .lib.fig_draw_multiplot import get_xylabels, draw_subplot
if FOUND_CARTOPY:
    from .lib.fig_draw_tracks import fig_draw_tracks


# - local functions --------------------------------

# - main function ----------------------------------
class MONplot:
    """
    Generate figure(s) for SRON onground calibration anlysis or
    inflight instrument monitoring

    Attributes
    ----------
    figname : str
       Name of the PDF or PNG file

    Methods
    -------
    close()
       Close PNG or (multipage) PDF document.
    set_caption(caption)
       Set caption of each page of the PDF.
    caption
       Return figure caption.
    set_cmap(cmap)
       Use alternative color-map for MONplot::draw_image.
    unset_cmap()
       Unset user supplied color-map, and use default color-map.
    cmap
       Return matplotlib colormap.
    set_institute(institute)
       Use the name of your institute as a signature.
    institute
       Return name of your institute.
    draw_signal(data, fig_info=None, side_panels='nanmedian', title=None,
                **kwargs)
       Display 2D array data as an image and averaged column/row signal plots.
    draw_quality(data, ref_data=None, side_panels='quality', fig_info=None,
                 title=None, **kwargs)
       Display pixel-quality 2D data as an image and column/row statistics.
    draw_trend(xds=None, hk_xds=None, fig_info=None, title=None, **kwargs)
       Display trends of measurement data and/or housekeeping data.
    draw_hist(data, data_sel=None, vrange=None, fig_info=None, title=None,
              **kwargs)
        Display data as a histogram.
    draw_qhist(xds, data_sel=None, density=True, fig_info=None, title=None)
       Display pixel-quality data as a histogram.
    draw_lplot(self, xdata, ydata, color=0, *, square=False, fig_info=None,
               title=None, **kwargs)
       Plot y versus x lines, maybe called multiple times to add lines.
    draw_multiplot(self, data_tuple, gridspec=None, fig_info=None, title=None,
                   **kwargs)
       Display multiple subplots on one page using matplotlib.gridspec.GridSpec.
    draw_tracks(lons, lats, icids, saa_region=None, fig_info=None, title=None)
       Display tracks of satellite on a world map using a Robinson projection.

    Notes
    -----
    ...
    """
    def __init__(self, figname, caption=None):
        """
        Initialize multi-page PDF document or a single-page PNG

        Parameters
        ----------
        figname :  str
           Name of PDF or PNG file (extension required)
        caption :  str
           Caption repeated on each page of the PDF

        Notes
        -----
        Currently, we have turned off the automatic offset notation of
        Matplotlib. Maybe this should be the default, which the user may
        override.
        """
        self.__cmap = None
        self.__caption = '' if caption is None else caption
        self.__institute = ''
        self.__mpl = None
        self.__pdf = None
        self.filename = figname
        if PurePath(figname).suffix.lower() != '.pdf':
            return

        self.__pdf = PdfPages(figname)

        # turn off Matplotlib's automatic offset notation
        mpl.rcParams['axes.formatter.useoffset'] = False

    def __repr__(self) -> None:
        pass

    def __close_this_page(self, fig) -> None:
        """
        Close current matplotlib figure or page in a PDF document
        """
        # add save figure
        if self.__pdf is None:
            plt.savefig(self.filename)
            plt.close(fig)
        else:
            self.__pdf.savefig()

    def close(self) -> None:
        """
        Close PNG or (multipage) PDF document
        """
        if self.__pdf is None:
            return

        # add PDF annotations
        doc = self.__pdf.infodict()
        if self.__caption is not None:
            doc['Title'] = self.__caption
        doc['Subject'] = \
            'Generated using https://github.com/rmvanhees/moniplot.git'
        if self.__institute == 'SRON':
            doc['Author'] = '(c) SRON Netherlands Institute for Space Research'
        elif self.__institute:
            doc['Author'] = f'(c) {self.__institute}'
        self.__pdf.close()
        plt.close('all')

    # --------------------------------------------------
    @property
    def caption(self) -> str:
        """
        Return figure caption
        """
        return self.__caption

    def set_caption(self, caption: str) -> None:
        """
        Set caption of each page of the PDF

        Parameter
        ---------
        institute :  str
           Provide abbreviation of the name of your institute to be used in
           the copyright statement in the main panel of the figures.
        """
        self.__caption = caption

    def __add_caption(self, fig):
        """
        Add figure caption
        """
        if not self.caption:
            return

        fig.suptitle(self.caption, fontsize='x-large',
                     position=(0.5, 1 - 0.3 / fig.get_figheight()))

    # --------------------------------------------------
    @property
    def cmap(self):
        """
        Return matplotlib colormap
        """
        return self.__cmap

    def set_cmap(self, cmap) -> None:
        """
        Use alternative color-map for MONplot::draw_image

        Parameter
        ---------
         cmap :  matplotlib color-map
        """
        self.__cmap = cmap

    def unset_cmap(self) -> None:
        """
        Unset user supplied color-map, and use default color-map
        """
        self.__cmap = None

    # --------------------------------------------------
    @property
    def institute(self) -> str:
        """
        Return name of your institute
        """
        return self.__institute

    def set_institute(self, institute: str) -> None:
        """
        Use the name of your institute as a signature

        Parameter
        ---------
        institute :  str
           Provide abbreviation of the name of your institute to be used in
           the copyright statement in the main panel of the figures.
        """
        self.__institute = institute

    # --------------------------------------------------
    def __add_copyright(self, axx) -> None:
        """
        Show value of institute as copyright in the lower right corner
        of the current figure.
        """
        if not self.institute:
            return

        axx.text(1, 0, rf' $\copyright$ {self.institute}',
                 horizontalalignment='right',
                 verticalalignment='bottom',
                 rotation='vertical', fontsize='xx-small',
                 transform=axx.transAxes)

    @staticmethod
    def __add_fig_box(fig, fig_info) -> None:
        """
        Add a box with meta information in the current figure

        Parameters
        ----------
        fig :  Matplotlib figure instance
        fig_info :  FIGinfo
           instance of pys5p.lib.plotlib.FIGinfo to be displayed
        """
        if fig_info is None or fig_info.location != 'above':
            return

        xpos = 1 - 0.4 / fig.get_figwidth()
        ypos = 1 - 0.25 / fig.get_figheight()

        fig.text(xpos, ypos, fig_info.as_str(),
                 fontsize='x-small', style='normal',
                 verticalalignment='top',
                 horizontalalignment='right',
                 multialignment='left',
                 bbox={'facecolor': 'white', 'pad': 5})

    # -------------------------
    def __draw_image__(self, xarr: xr.DataArray, side_panels: str,
                       fig_info: FIGinfo, title: str) -> None:
        """
        Does the actual drawing of the image data for the public methods
        'draw_signal' and 'draw_quality'.
        """
        def add_fig_box() -> None:
            """
            Add a box with meta information in the current figure
            """
            if fig_info is None:
                return

            if fig_info.location == 'above':
                if aspect <= 2:
                    halign = 'center'
                    fontsize = 'x-small'
                else:
                    halign = 'right'
                    fontsize = 'xx-small' if len(fig_info) > 6 else 'x-small'

                axx_c.text(0 if aspect == 2 else 1,
                           1.04 + (aspect-1) * 0.0075,
                           fig_info.as_str(), fontsize=fontsize,
                           transform=axx_c.transAxes,
                           multialignment='left',
                           verticalalignment='bottom',
                           horizontalalignment=halign,
                           bbox={'facecolor': 'white', 'pad': 4})
                return

            if fig_info.location == 'below':
                fontsize = 'xx-small' if aspect in (3, 4) else 'x-small'
                axx_c.text(0.125 + (aspect-1) * 0.2,
                           -0.03 - (aspect-1) * 0.005,
                           fig_info.as_str(), fontsize=fontsize,
                           transform=axx_c.transAxes,
                           multialignment='left',
                           verticalalignment='top',
                           horizontalalignment='left',
                           bbox={'facecolor': 'white', 'pad': 4})

        # aspect of image data
        aspect = min(4, max(1, int(round(xarr.shape[1] / xarr.shape[0]))))

        # select figure attributes
        attrs = {1: {'figsize': (10, 8),
                     'w_ratios': (1., 7., 0.5, 1.5),
                     'h_ratios': (7., 1.)},                  # 7 x 7
                 2: {'figsize': (13, 6.25),
                     'w_ratios': (1., 10., 0.5, 1.5),
                     'h_ratios': (5., 1.)},                  # 10 x 5
                 3: {'figsize': (15, 5.375),
                     'w_ratios': (1., 12., 0.5, 1.5),
                     'h_ratios': (4., 1.)},                  # 12 x 4
                 4: {'figsize': (17, 5.125),
                     'w_ratios': (1., 14., 0.5, 1.5),
                     'h_ratios': (3.5, 1.)}}.get(aspect)     # 14 x 3.5

        # define matplotlib figure
        fig = plt.figure(figsize=attrs['figsize'])
        if self.caption:
            fig.suptitle(self.caption, fontsize='x-large',
                         position=(0.5, 1 - 0.4 / fig.get_figheight()))

        # Define a grid layout to place subplots within the figure.
        # - gspec[0, 1] is reserved for the image
        # - gspec[1, 1] is reserved for the x-panel
        # - gspec[0, 0] is reserved for the y-panel
        # - gspec[0, 2] is reserved for the colorbar
        # - gspec[1, 2] is used to pace the small fig_info box (max 6/7 lines)
        gspec = fig.add_gridspec(2, 4,
                                 left=.135 + .005 * (aspect-1),
                                 right=.9 - .005 * (aspect-1),
                                 top=.865 - .025 * (aspect-1),
                                 bottom=.115 + .01 * (aspect-1),
                                 wspace=0.1 / max(2, aspect-1),
                                 hspace=0.05,
                                 width_ratios=attrs['w_ratios'],
                                 height_ratios=attrs['h_ratios'])

        # add image panel and draw image
        axx = fig.add_subplot(gspec[0, 1])
        if xarr.attrs['_zscale'] == 'quality':
            img = axx.imshow(xarr.values, norm=xarr.attrs['_znorm'],
                             aspect='auto', cmap=xarr.attrs['_cmap'],
                             interpolation='none', origin='lower')
        else:
            cmap = self.cmap if self.cmap else xarr.attrs['_cmap']
            img = axx.imshow(xarr.values, norm=xarr.attrs['_znorm'],
                             aspect='auto', cmap=cmap,
                             interpolation='none', origin='lower')

        # add title to image panel
        if title is not None:
            axx.set_title(title)
        elif 'long_name' in xarr.attrs:
            axx.set_title(xarr.attrs['long_name'])
        self.__add_copyright(axx)
        # axx.grid(True)

        # add side panels
        if side_panels == 'none':
            adjust_img_ticks(axx, xarr)
            axx.set_xlabel(xarr.dims[1])
            axx.set_ylabel(xarr.dims[0])
        else:
            for xtl in axx.get_xticklabels():
                xtl.set_visible(False)
            for ytl in axx.get_yticklabels():
                ytl.set_visible(False)
            axx_p = {'X': fig.add_subplot(gspec[1, 1], sharex=axx),
                     'Y': fig.add_subplot(gspec[0, 0], sharey=axx)}
            fig_draw_panels(axx_p, xarr, side_panels)
            axx_p['X'].set_xlabel(xarr.dims[1])
            axx_p['Y'].set_ylabel(xarr.dims[0])

        # add colorbar
        if xarr.attrs['_zscale'] == 'quality':
            axx_c = fig.add_subplot(gspec[0, 2])
            bounds = xarr.attrs['flag_values']
            mbounds = [(bounds[ii+1] + bounds[ii]) / 2
                       for ii in range(len(bounds)-1)]
            _ = plt.colorbar(img, cax=axx_c, ticks=mbounds, boundaries=bounds)
            axx_c.tick_params(axis='y', which='both', length=0)
            axx_c.set_yticklabels(xarr.attrs['flag_meanings'])
        else:
            axx_c = fig.add_subplot(gspec[0, 2])
            _ = plt.colorbar(img, cax=axx_c, label=xarr.attrs['_zlabel'])

        # add annotation and save the figure
        add_fig_box()
        self.__close_this_page(fig)

    # --------------------------------------------------
    def draw_signal(self, data, *, fig_info=None, side_panels='nanmedian',
                    title=None, **kwargs) -> None:
        """
        Display 2D array data as image and averaged column/row signal plots

        Parameters
        ----------
        data :  numpy.ndarray or xarray.DataArray
           Object holding measurement data and attributes
        fig_info :  FIGinfo, default=None
           OrderedDict holding meta-data to be displayed in the figure
        side_panels :  str, default='nanmedian'
           Show image row and column statistics in two side panels.
           Use 'none' when you do not want the side panels.
           Other valid values are: 'median', 'nanmedian', 'mean', 'nanmean',
           'quality', 'std' and 'nanstd'.
        title :  str, default=None
           Title of this figure (matplotlib: Axis.set_title)
        **kwargs :   other keywords
           Pass keyword arguments: 'zscale', 'vperc' or 'vrange'
           to moniplot.lib.fig_draw_image.fig_data_to_xarr()

        The information provided in the parameter 'fig_info' will be displayed
        in a text box. In addition, we display the creation date and the data
        (biweight) median & spread.

        xarray attributes
        -----------------
        Attributes starting with an underscore are added by fig_data_to_xarr

        long_name :  used as the title of the main panel when parameter 'title'
            is not defined.
        _cmap :  contains the matplotlib colormap
        _zlabel :  contains the label of the color bar
        _znorm :  matplotlib class to normalize the data between zero and one.
        _zscale :  scaling of the data values: linear, log, diff, ratio, ...
        _zunits :  adjusted units of the data

        Examples
        --------
        Create a PDF document 'test.pdf' and add figure of dataset 'img'
        (np.ndarray or xr.DataArray) with side-panels and title

        >>> plot = MONplot('test.pdf', caption='my caption')
        >>> plot.set_institute('SRON')
        >>> plot.draw_signal(img, title='my title')

        Add the same figure without side-panels

        >>> plot.draw_signal(img, side_panels='none', title='my title')

        Add a figure using a fixed data-range that the colormap covers

        >>> plot.draw_signal(img1, title='my title', vrange=[zmin, zmax])

        Add a figure where img2 = img - img_ref

        >>> plot.draw_signal(img2, title='my title', zscale='diff')

        Add a figure where img2 = img / img_ref

        >>> plot.draw_signal(img2, title='my title', zscale='ratio')

        Finalize the PDF file

        >>> plot.close()

        """
        # convert, if necessary, input data to xarray.DataArray
        if isinstance(data, xr.DataArray) and '_zscale' in data.attrs:
            xarr = data.copy()
        else:
            xarr = fig_data_to_xarr(data, **kwargs)

        # add data statistics to fig_info
        if fig_info is None:
            fig_info = FIGinfo()

        median, spread = biweight(xarr.values, spread=True)
        if xarr.attrs['_zunits'] is None or xarr.attrs['_zunits'] == '1':
            fig_info.add('median', median, '{:.5g}')
            fig_info.add('spread', spread, '{:.5g}')
        else:
            fig_info.add('median', (median, xarr.attrs['_zunits']), '{:.5g} {}')
            fig_info.add('spread', (spread, xarr.attrs['_zunits']), '{:.5g} {}')

        # draw actual image
        self.__draw_image__(xarr, side_panels, fig_info, title)

    # --------------------------------------------------
    def draw_quality(self, data, ref_data=None, *, side_panels='quality',
                     fig_info=None, title=None, **kwargs) -> None:
        """
        Display pixel-quality 2D array data as image and column/row statistics

        Parameters
        ----------
        data :  numpy.ndarray or xarray.DataArray
           Object holding measurement data and attributes
        ref_data :  numpy.ndarray, default=None
           Numpy array holding reference data, for example pixel quality
           reference map taken from the CKD. Shown are the changes with
           respect to the reference data.
        fig_info :  FIGinfo, default=None
           OrderedDict holding meta-data to be displayed in the figure
        side_panels :  str, default='quality'
           Show image row and column statistics in two side panels.
           Use 'none' when you do not want the side panels.
        title :  str, default=None
           Title of this figure (matplotlib: Axis.set_title)
        **kwargs :   other keywords
           Pass keyword arguments: 'data_sel', 'thres_worst', 'thres_bad'
           or 'qlabels' to moniplot.lib.fig_draw_image.fig_qdata_to_xarr()

        xarray attributes
        -----------------
        Attributes starting with an underscore are added by fig_qdata_to_xarr

        long_name :  used as the title of the main panel when parameter 'title'
            is not defined.
        flag_values :  values of the flags used to qualify the pixel quality
        flag_meanings :  description of the flag values
        thres_bad :  threshold between good and bad
        thres_worst :  threshold between bad and worst
        _cmap :  contains the matplotlib colormap.
        _zscale :  'quality'.
        _znorm :  matplotlib class to normalize the data between zero and one.

        Notes
        -----
        The quality ranking labels are ['unusable', 'worst', 'bad', 'good'],
        in case nor reference dataset is provided. Where:
        - 'unusable'  : pixels outside the illuminated region
        - 'worst'     : 0 <= value < thres_worst
        - 'bad'       : 0 <= value < thres_bad
        - 'good'      : thres_bad <= value <= 1
        Otherwise the labels for quality ranking indicate which pixels have
        changed w.r.t. reference. The labels are:
        - 'unusable'  : pixels outside the illuminated region
        - 'worst'     : from good or bad to worst
        - 'bad'       : from good to bad
        - 'good'      : from any rank to good
        - 'unchanged' : no change in rank

        The information provided in the parameter 'fig_info' will be displayed
        in a small box. Where creation date and statistics on the number of
        bad and worst pixels are displayed.

        Examples
        --------
        Create a PDF document 'test.pdf' and add figure of dataset 'img'
        (np.ndarray or xr.DataArray) with side-panels and title

        >>> plot = MONplot('test.pdf', caption='my caption', institute='SRON')
        >>> plot.draw_quality(img, title='my title')

        Add the same figure without side-panels

        >>> plot.draw_quality(img, side_panels='none', title='my title')

        Add a figure where img_ref is a quality map from early in the mission

        >>> plot.draw_quality(img, img_ref, title='my title')

        Finalize the PDF file

        >>> plot.close()
        """
        # convert, if necessary, input data to xarray.DataArray
        if isinstance(data, xr.DataArray) and '_zscale' in data.attrs:
            xarr = data
        else:
            xarr = fig_qdata_to_xarr(data, ref_data, **kwargs)

        # add statistics on data quality to fig_info
        if fig_info is None:
            fig_info = FIGinfo()

        if ref_data is None:
            fig_info.add(
                f'{xarr.attrs["flag_meanings"][2]}'
                f' (quality < {xarr.attrs["thres_bad"]})',
                np.sum((xarr.values == 1) | (xarr.values == 2)))
            fig_info.add(
                f'{xarr.attrs["flag_meanings"][1]}'
                f' (quality < {xarr.attrs["thres_worst"]})',
                np.sum(xarr.values == 1))
        else:
            fig_info.add(xarr.attrs['flag_meanings'][3],
                         np.sum(xarr.values == 4))
            fig_info.add(xarr.attrs['flag_meanings'][2],
                         np.sum(xarr.values == 2))
            fig_info.add(xarr.attrs['flag_meanings'][1],
                         np.sum(xarr.values == 1))

        # draw actual image
        self.__draw_image__(xarr, side_panels, fig_info, title)

    # --------------------------------------------------
    def draw_trend(self, xds=None, hk_xds=None, *,
                   fig_info=None, title=None, **kwargs) -> None:
        """
        Display trends of measurement data and/or housekeeping data

        Parameters
        ----------
        xds :  xarray.Dataset, optional
           Object holding measurement data and attributes
        hk_xds :  xarray.Dataset, optional
           Object holding housekeeping data and attributes
        fig_info :  FIGinfo, optional
           OrderedDict holding meta-data to be displayed in the figure
        title :  str, optional
           Title of this figure (matplotlib: Axis.set_title)
        **kwargs :   other keywords
           Pass keyword arguments: 'vperc' or 'vrange_last_orbits'
           to moniplot.lib.fig_draw_trend.add_hk_subplot()

        xarray attributes
        -----------------
        long_name :  used as the title of the main panel when parameter 'title'
            is not defined.
        units :  units of the data

        Examples
        --------
        Create a PDF document 'test.pdf' and add figure of dataset 'xds'
        (np.ndarray or xr.DataArray) with a title. The dataset 'xds' may
        contain multiple DataArrays with a common X-coordinate. Each DataArray
        will be displayed in a seperate sub-panel.

        >>> plot = MONplot('test.pdf', caption='my caption', institute='SRON')
        >>> plot.draw_trend(xds, hk_xds=None, title='my title')

        Add a figure with the same Dataset 'xds' and a few trends of
        housekeeping data (again each parameter in a seperate DataArray with
        with a common X-coordinate).

        >>> plot.draw_trend(xds, hk_xds, title='my title')

        Finalize the PDF file

        >>> plot.close()

        """
        if xds is None and hk_xds is None:
            raise ValueError('both xds and hk_xds are None')
        if xds is not None and not isinstance(xds, xr.Dataset):
            raise ValueError('xds should be and xarray Dataset object')
        if hk_xds is not None and not isinstance(hk_xds, xr.Dataset):
            raise ValueError('hk_xds should be and xarray Dataset object')

        if fig_info is None:
            fig_info = FIGinfo()

        # determine npanels from xarray Dataset
        npanels = len(xds.data_vars) if xds is not None else 0
        npanels += len(hk_xds.data_vars) if hk_xds is not None else 0

        # initialize matplotlib using 'subplots'
        figsize = (10., 1 + (npanels + 1) * 1.5)
        fig, axarr = plt.subplots(npanels, sharex=True, figsize=figsize)
        if npanels == 1:
            axarr = [axarr]
        margin = min(1. / (1.65 * (npanels + 1)), .25)
        fig.subplots_adjust(bottom=margin, top=1-margin, hspace=0.05)

        # add a centered suptitle to the figure
        self.__add_caption(fig)

        # add title to image panel
        if title is not None:
            axarr[0].set_title(title)

        # add figures with trend data
        ipanel = 0
        if xds is not None:
            xlabel = 'orbit' if 'orbit' in xds.coords else 'time [hours]'
            for name in xds.data_vars:
                add_subplot(axarr[ipanel], xds[name])
                ipanel += 1

        if hk_xds is not None:
            xlabel = 'orbit' if 'orbit' in hk_xds.coords else 'time [hours]'
            for name in hk_xds.data_vars:
                add_hk_subplot(axarr[ipanel], hk_xds[name], **kwargs)
                ipanel += 1

        # finally add a label for the X-coordinate
        axarr[-1].set_xlabel(xlabel)

        # add annotation and save the figure
        self.__add_copyright(axarr[-1])
        self.__add_fig_box(fig, fig_info)
        self.__close_this_page(fig)

    # --------------------------------------------------
    def draw_hist(self, data, data_sel=None, vrange=None,
                  fig_info=None, title=None, **kwargs) -> None:
        """
        Display data as histograms.

        Parameters
        ----------
        data :  numpy.ndarray or xarray.DataArray
           Object holding measurement data and attributes
        data_sel :  mask or index tuples for arrays, optional
           Select a region on the detector by fancy indexing (using a
           boolean/interger arrays), or using index tuples for arrays
           (generated with numpy.s_).
        vrange :  list, default=[data.min(), data.max()]
           The lower and upper range of the bins.
           Note data will also be clipped according to this range.
        fig_info :  FIGinfo, optional
           OrderedDict holding meta-data to be displayed in the figure
        title :  str, optional
           Title of this figure (matplotlib: Axis.set_title)
        **kwargs :   other keywords
           Pass the following keyword arguments to matplotlib.pyplot.hist:
              'bins', 'density' or 'log'.
           Note that keywords: 'histtype', 'color', 'linewidth' and 'fill'
           are predefined.

        xarray attributes
        -----------------
        long_name :  used as the title of the main panel when parameter 'title'
            is not defined.
        units :  units of the data

        Examples
        --------
        Create a PDF document 'test.pdf' and add figure with a histogram
        of array 'data' (np.ndarray or xr.DataArray). And a second page
        with also a histogram where data is a xarray with attribute
        'long_name' then the title is
        f'Histograms of {xarr.attrs["long_name"]} values'

        The array will be flattend the the histogram is created with
        matplotlib.pyplot.hist.

        >>> plot = MONplot('test.pdf', caption='my caption')
        >>> plot.set_institute('SRON')
        >>> plot.draw_hist(data, title='my title')
        >>> plot.draw_hist(xarr)
        >>> plot.close()
        """
        long_name = ''
        zunits = '1'
        if isinstance(data, xr.DataArray):
            if data_sel is None:
                values = data.values.reshape(-1)
            else:
                values = data.values[data_sel].reshape(-1)
            if 'long_name' in data.attrs:
                long_name = data.attrs['long_name']
            if 'units' in data.attrs:
                zunits = data.attrs['units']
        else:
            if data_sel is None:
                values = data.reshape(-1)
            else:
                values = data[data_sel].reshape(-1)

        # add data statistics to fig_info
        if fig_info is None:
            fig_info = FIGinfo()

        median, spread = biweight(values, spread=True)
        if zunits == '1':
            fig_info.add('median', median, '{:.5g}')
            fig_info.add('spread', spread, '{:.5g}')
        else:
            fig_info.add('median', (median, zunits), '{:.5g} {}')
            fig_info.add('spread', (spread, zunits), '{:.5g} {}')

        # create figure
        fig, axx = plt.subplots(1, figsize=(8, 7))

        # add a centered suptitle to the figure
        self.__add_caption(fig)

        # add title to image panel
        if title is None:
            title = f'Histograms of {long_name} values'
        axx.set_title(title)

        # add histogram
        if vrange is not None:
            values = np.clip(values, vrange[0], vrange[1])
        # Edgecolor is tol_cset('bright').blue
        if 'bins' in kwargs and kwargs['bins'] > 24:
            axx.hist(values, range=vrange, histtype='step',
                     edgecolor='#4477AA', facecolor='#77AADD',
                     fill=True, linewidth=1.5, **kwargs)
            axx.grid(which='major', color='#AAAAAA', linestyle='--')
        else:
            axx.hist(values, range=vrange, histtype='bar',
                     edgecolor='#4477AA', facecolor='#77AADD',
                     linewidth=1.5, **kwargs)
            axx.grid(which='major', axis='y', color='#AAAAAA', linestyle='--')
        axx.set_xlabel(long_name if long_name else 'value')
        if 'density' in kwargs and kwargs['density']:
            axx.set_ylabel('density')
        else:
            axx.set_ylabel('count')

        # add annotation and save the figure
        self.__add_copyright(axx)
        self.__add_fig_box(fig, fig_info)
        self.__close_this_page(fig)

    # --------------------------------------------------
    def draw_qhist(self, xds, data_sel=None, density=True,
                   fig_info=None, title=None) -> None:
        """
        Display pixel-quality data as histograms.

        Parameters
        ----------
        xds :  xarray.Dataset
           Object holding measurement data and attributes
        data_sel :  mask or index tuples for arrays, optional
           Select a region on the detector by fancy indexing (using a
           boolean/interger arrays), or using index tuples for arrays
           (generated with numpy.s_).
        density : bool, default=True
           If True, draw and return a probability density: each bin will
           display the bin's raw count divided by the total number of counts
           and the bin width (see matplotlib.pyplot.hist).
        fig_info :  FIGinfo, optional
           OrderedDict holding meta-data to be displayed in the figure
        title :  str, optional
           Title of this figure (matplotlib: Axis.set_title)

        xarray attributes
        -----------------
        long_name :  used as the title of the main panel when parameter 'title'
            is not defined.
        units :  units of the data

        Examples
        --------
        Create a PDF document 'test.pdf' and add figure of dataset 'xds'
        (np.ndarray or xr.DataArray) with a title. The dataset 'xds' may
        contain multiple DataArrays with a common X-coordinate. Each DataArray
        will be displayed in a seperate sub-panel.

        >>> plot = MONplot('test.pdf', caption='my caption', institute='SRON')
        >>> plot.draw_qhist(xds, title='my title')
        >>> plot.close()

        """
        if not isinstance(xds, xr.Dataset):
            raise ValueError('xds should be and xarray Dataset object')

        if fig_info is None:
            fig_info = FIGinfo()

        # determine npanels from xarray Dataset
        npanels = len(xds.data_vars)

        # initialize matplotlib using 'subplots'
        figsize = (10., 1 + (npanels + 1) * 1.65)
        fig, axarr = plt.subplots(npanels, sharex=True, figsize=figsize)
        if npanels == 1:
            axarr = [axarr]
        margin = min(1. / (1.8 * (npanels + 1)), .25)
        fig.subplots_adjust(bottom=margin, top=1-margin, hspace=0.02)

        # add a centered suptitle to the figure
        self.__add_caption(fig)

        # add title to image panel
        if title is None:
            title = 'Histograms of pixel-quality'
        axarr[0].set_title(title)

        # add figures with histograms
        for ii, (key, xda) in enumerate(xds.data_vars.items()):
            if data_sel is None:
                qdata = xda.values.reshape(-1)
            else:
                qdata = xda.values[data_sel].reshape(-1)
            label = xda.attrs['long_name'] if 'long_name' in xda.attrs else key
            fig_draw_qhist(axarr[ii], qdata, label, density)

        # finally add a label for the X-coordinate
        axarr[-1].set_xlabel('pixel quality')

        # add annotation and save the figure
        self.__add_copyright(axarr[-1])
        self.__add_fig_box(fig, fig_info)
        self.__close_this_page(fig)

    # --------------------------------------------------
    def draw_lplot(self, xdata, ydata, color=0, *, square=False,
                   fig_info=None, title=None, **kwargs) -> None:
        """
        Plot y versus x lines, maybe called multiple times to add lines.
        Figure is closed when called with x is None.

        Parameters
        ----------
        xdata :  ndarray
           [add line] X data, [close figure] when xdata is None
        ydata :  ndarray
           [add line] Y data
        square :  bool
           [first call, only] create a square figure, independent of
           number of data-points.
        color :  integer, default=0
           [add line] Index to color in tol_colors.tol_cset('bright')
        fig_info  :  FIGinfo, optional
           [close figure] Meta-data to be displayed in the figure
        title :  str, default=None
           [close figure] Title of this figure (matplotlib: Axis.set_title)
        **kwargs :   other keywords
           [add line] Keywords are passed to mpl.pyplot.plot()
           [close figure] Kewords are passed to appropriate mpl.Axes method
           [close figure] keyword 'text' can be used to add addition text in
               the upper left corner.

        Examples
        --------
        General example:
        >>> plot = MONplot(fig_name)
        >>> for ii, ix, iy in enumerate(data_of_each_line):
        >>>    plot.draw_lplot(ix, iy, color=ii, label=mylabel[ii],
        >>>                    marker='o', linestyle='None')
        >>> plot.draw_lplot(None, None, xlim=[0, 0.5], ylim=[-10, 10],
        >>>                 xlabel=my_xlabel, ylabel=my_ylabel)
        >>> plot.close()

        Using a time-axis:
        >>> from datetime import datetime, timedelta
        >>> tt0 = (datetime(year=2020, month=10, day=1)
        >>>        + timedelta(seconds=sec_in_day))
        >>> tt = [tt0 + iy * t_step for iy in range(yy.size)]
        >>> plot = MONplot(fig_name)
        >>> plot.draw_lplot(tt, yy, color=1, label=mylabel,
        >>>                 marker='o', linestyle='None')
        >>> plot.draw_line(None, None, ylim=[-10, 10],
        >>>                xlabel=my_xlabel, ylabel=my_ylabel)
        >>> plot.close()
        """
        if xdata is None:
            if self.__mpl is None:
                raise ValueError('No plot defined and no data provided')
            fig = self.__mpl['fig']
            axx = self.__mpl['axx']

            if fig_info is None:
                fig_info = FIGinfo()

            if 'text' in kwargs:
                axx.text(0.05, 0.985, kwargs.pop('text'),
                         transform=axx.transAxes,
                         fontsize='small', verticalalignment='top',
                         bbox=dict(boxstyle='round', facecolor='#FFFFFF',
                                   edgecolor='#BBBBBB', alpha=0.5))

            close_draw_lplot(axx, self.__mpl['time_axis'], title, **kwargs)

            # add annotation and save the figure
            self.__add_copyright(axx)
            self.__add_fig_box(fig, fig_info)
            self.__close_this_page(fig)
            self.__mpl = None
            return

        # initialize figure
        if self.__mpl is None:
            if square:
                figsize = (9, 9)
            else:
                figsize = {0: (8, 7),
                           1: (10, 7),
                           2: (12, 7)}.get(len(xdata) // 256, (14, 8))

            self.__mpl = dict(zip(('fig', 'axx'),
                                  plt.subplots(1, figsize=figsize)))
            self.__mpl['time_axis'] = isinstance(xdata[0], datetime)

            # add a centered suptitle to the figure
            self.__add_caption(self.__mpl['fig'])

        # draw line in figure
        fig_draw_lplot(self.__mpl['axx'], xdata, ydata, color, **kwargs)

    # --------------------------------------------------
    def draw_multiplot(self, data_tuple: tuple, gridspec=None, *,
                       fig_info=None, title=None, **kwargs) -> None:
        """
        Display multiple subplots on one page using
        matplotlib.gridspec.GridSpec

        Parameters
        ----------
        data_tuple :  tuple with nparray, xarray.DataArray or xarray.Dataset
           One dataset per subplot
        gridspec :  matplotlib.gridspec.GridSpec, optional
           Instance of matplotlib.gridspec.GridSpec
        fig_info  :  FIGinfo, optional
           Meta-data to be displayed in the figure
        title :  str, default=None
           Title of this figure (matplotlib: Axis.set_title)
           Ignored when data is a xarray data structure
        **kwargs :   other keywords
           Keywords are passed to mpl.pyplot.plot().
           Ignored when data is a xarray data structure

        xarray attributes
        -----------------
        long_name :  used as the title of the main panel when parameter 'title'
            is not defined.
        units :  units of the data
        _plot :  dictionary with parameters for matplotlib.pyplot.plot
        _title :  title of the subplot (matplotlib: Axis.set_title)
        _text :  text shown in textbox placed in the upper left corner
        _yscale :  y-axis scale type, default 'linear'
        _xlim :  range of the x-axis
        _ylim :  range of the y-axis

        Notes
        -----

        Examples
        --------
        Show two numpy arrays:
        >>> data_tuple = (ndarray1, ndarray2)
        >>> plot = MONplot(fig_name)
        >>> plot.draw_multiplot(data_tuple: tuple, title='my title',
        >>>                     marker='o', linestyle='', color='r')
        >>> plot.close()
        """
        # generate figure using contrained layout
        fig = plt.figure(figsize=(10, 10))

        # define grid layout to place subplots within a figure
        if gridspec is None:
            geometry = {1: (1, 1),
                        2: (2, 1),
                        3: (3, 1),
                        4: (2, 2)}.get(len(data_tuple))
            gridspec = GridSpec(*geometry, figure=fig)
        else:
            if len(data_tuple) > gridspec.nrows * gridspec.ncols:
                raise RuntimeError('grid too small for number of datasets')

        # determine xylabels
        xylabels = get_xylabels(gridspec, data_tuple)

        # add a centered suptitle to the figure
        self.__add_caption(fig)

        # add subplots, cycle the DataArrays of the Dataset
        data_iter = iter(data_tuple)
        for iy in range(gridspec.nrows):
            for ix in range(gridspec.ncols):
                axx = fig.add_subplot(gridspec[iy, ix])
                axx.grid(True)

                data = next(data_iter)
                if isinstance(data, np.ndarray):
                    if ix == iy == 0 and title is not None:
                        axx.set_title(title)
                    axx.plot(np.arange(data.size), data, **kwargs)
                elif isinstance(data, xr.DataArray):
                    draw_subplot(axx, data, xylabels[iy, ix, :])
                else:
                    for name in data.data_vars:
                        draw_subplot(axx, data[name], xylabels[iy, ix, :])

        # add annotation and save the figure
        self.__add_copyright(axx)
        if fig_info is None:
            fig_info = FIGinfo()
        self.__add_fig_box(fig, fig_info)
        self.__close_this_page(fig)

    # --------------------------------------------------
    def draw_tracks(self, lons, lats, icids, *, saa_region=None,
                    fig_info=None, title=None) -> None:
        """
        Display tracks of satellite on a world map using a Robinson projection

        Parameters
        ----------
        lons :  (N, 2) array-like
           Longitude coordinates at start and end of measurement
        lats :  (N, 2) array-like
           Latitude coordinates at start and end of measurement
        icids :  (N) array-like
           ICID of measurements per (lon, lat)
        saa_region :  (N, 2) array-like, optional
           The coordinates of the vertices. When defined, then show SAA region
           as a matplotlib polygon patch.
        fig_info :  FIGinfo, optional
           OrderedDict holding meta-data to be displayed in the figure

        The information provided in the parameter 'fig_info' will be displayed
        in a small box.
        """
        if not FOUND_CARTOPY:
            raise RuntimeError('You need Cartopy to run this function')

        if fig_info is None:
            fig_info = FIGinfo()

        # define plot layout
        # pylint: disable=abstract-class-instantiated
        myproj = {'projection': ccrs.Robinson(central_longitude=11.5)}
        fig, axx = plt.subplots(figsize=(12.85, 6), subplot_kw=myproj)

        # add a centered suptitle of the Figure
        self.__add_caption(fig)

        # add title to image panel
        if title is not None:
            axx.set_title(title)

        # draw tracks of satellite
        fig_draw_tracks(axx, lons, lats, icids, saa_region)

        # finalize figure
        self.__add_copyright(axx)
        self.__add_fig_box(fig, fig_info)
        self.__close_this_page(fig)



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

Python implementation of the Tukey's biweight algorithm.

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
import gc
import warnings

import numpy as np


# ----- local functions -------------------------
def __biweight(data, spread=False):
    """
    Calculate biweight parameters for the whole dataset
    """
    # calculate median and median absolute difference
    if np.isnan(data).any():
        biweight_median = np.nanmedian(data)
        delta = data - biweight_median
        delta_median = np.nanmedian(np.abs(delta))
    else:
        biweight_median = np.median(data)
        delta = data - biweight_median
        delta_median = np.median(np.abs(delta))
    if delta_median == 0.:
        if spread:
            return (biweight_median, 0.)
        return biweight_median

    # calculate biweight median
    wmx = np.clip(1 - (delta / (6 * delta_median)) ** 2, 0, None) ** 2
    biweight_median += np.nansum(wmx * delta) / np.nansum(wmx)
    del wmx
    gc.collect()
    if not spread:
        return biweight_median

    # calculate biweight variance
    umn = np.clip((delta / (9 * delta_median)) ** 2, None, 1)
    biweight_var = np.nansum(delta ** 2 * (1 - umn) ** 4)
    del delta
    gc.collect()
    biweight_var /= np.nansum((1 - umn) * (1 - 5 * umn)) ** 2
    biweight_var *= np.sum(np.isfinite(data))

    return (biweight_median, np.sqrt(biweight_var))


def __biweight_axis(data, axis, spread=False):
    """
    Calculate biweight parameters, along a given axis

    Notes
    -----
    Data should not contain any invalid value
    """
    # calculate median and median absolute difference
    if np.isnan(data).any():
        all_nan = np.isnan(data).all(axis=axis)
        biweight_median = np.nanmedian(data, axis=axis, keepdims=True)
        delta = data - biweight_median
        delta_median = np.nanmedian(np.abs(delta), axis=axis, keepdims=True)
        _mm = delta_median != 0.
        delta_median[~_mm] = np.nan
        _mm = np.squeeze(_mm) & ~all_nan
    else:
        biweight_median = np.median(data, axis=axis, keepdims=True)
        delta = data - biweight_median
        delta_median = np.median(np.abs(delta), axis=axis, keepdims=True)
        _mm = delta_median != 0.
        delta_median[~_mm] = np.nan
        _mm = np.squeeze(_mm)
    biweight_median = np.squeeze(biweight_median)

    # calculate biweight median
    wmx = np.clip(1 - (delta / (6 * delta_median)) ** 2, 0, None) ** 2
    biweight_median[_mm] += \
        np.nansum(wmx * delta, axis=axis)[_mm] / np.nansum(wmx, axis=axis)[_mm]
    del wmx
    gc.collect()
    if not spread:
        return biweight_median

    # calculate biweight variance
    umn = np.clip((delta / (9 * delta_median)) ** 2, None, 1)
    del delta_median
    biweight_var = np.nansum(delta ** 2 * (1 - umn) ** 4, axis=axis)
    del delta
    gc.collect()
    biweight_var[_mm] /= \
        np.nansum((1 - umn) * (1 - 5 * umn), axis=axis)[_mm] ** 2
    biweight_var[_mm] *= np.sum(np.isfinite(data), axis=axis)[_mm]

    return (biweight_median, np.sqrt(biweight_var))


# ----- main function -------------------------
def biweight(data, axis=None, spread=False):
    """
    Calculate Tukey's biweight.
    Implementation based on Eqn. 7.6 and 7.7 in the SWIR OCAL ATBD.

    Parameters
    ----------
    data   :   array_like
       input array
    axis   :   int, optional
       axis along which the biweight medians are computed.
        - Note that axis will be ignored when data is a 1-D array.
    spread :   bool, optional
       if True, then return also the biweight spread.

    Returns
    -------
    out    :   ndarray or tuple
       biweight median and biweight spread if parameter "spread" is True
    """
    if axis is None or data.ndim == 1:
        mask = np.isfinite(data)
        if np.all(~mask):
            if spread:
                return (np.nan, 0.)
            return np.nan
        return __biweight(data, spread)

    # only a single axis!
    if not isinstance(axis, int):
        raise TypeError('axis not an integer')
    if not 0 <= axis < data.ndim:
        raise ValueError('axis out-of-range')

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
        return __biweight_axis(data, axis, spread)



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

This module contains the definition of the class FIGinfo

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
from copy import deepcopy
from datetime import datetime


# - main function -------------------------
class FIGinfo:
    """
    The figure information constists of [key, value] combinations which are
    to be displayed upper-right corner of the figure.

    Attributes
    ----------
    location : str
       Location to draw the fig_info box: 'above' (default), 'none'
    fig_info : dict
       Dictionary holding the information for the fig_info box

    Methods
    -------
    add(key, value, fmt='{}')
       Extent fig_info with a new line.
    as_str()
       Return figure information as one long string.
    copy()
       Return a deep copy of the current object.
    set_location(loc)
       Set the location of the fig_info box.

    Notes
    -----
    The box with the figure information can only hold a limited number of keys:
      'above' :  The figure information is displayed in a small box. This box
                 grows with the number of lines and will overlap with the main
                 image or its colorbar at about 7+ entries.
    """
    def __init__(self, loc='above', info_dict=None) -> None:
        """
        Create FIGinfo instance to hold information on the current plot

        Parameters
        ----------
        loc :  str, default='above'
           Location to draw the fig_info box: 'above' (default), 'none'
        info_dict :  dict, optional
           Dictionary holding information to be displayed in the fig_info box
        """
        self.fig_info = {} if info_dict is None else info_dict
        self.set_location(loc)

    def __bool__(self) -> bool:
        return bool(self.fig_info)

    def __len__(self) -> int:
        return len(self.fig_info)

    def copy(self):
        """
        Return a deep copy of the current object
        """
        return deepcopy(self)

    def set_location(self, loc: str) -> None:
        """
        Set the location of the fig_info box

        Parameters
        ----------
        loc : str
          Location of the fig_info box
        """
        if loc not in ('above', 'none'):
            raise KeyError("location should be: 'above' or 'none'")

        self.location = loc

    def add(self, key: str, value, fmt='{}') -> None:
        """
        Extent fig_info with a new line

        Parameters
        ----------
        key : str
          Name of the fig_info key.
        value : Any python variable
          Value of the fig_info key. A tuple will be formatted as *value.
        fmt : str, default='{}'
          Convert value to a string, using the string format method.
        """
        if isinstance(value, tuple):
            self.fig_info[key] = fmt.format(*value)
        else:
            self.fig_info[key] = fmt.format(value)

    def as_str(self) -> str:
        """
        Return figure information as one long string
        """
        info_str = ''
        for key, value in self.fig_info.items():
            info_str += f'{key} : {value}\n'

        # add timestamp
        res = datetime.utcnow().isoformat(timespec='seconds')
        info_str += f'created : {res}'

        return info_str



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

Defines functions fig_draw_lplot and close_draw_lplot
used by MONplot::draw_lplot()

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter

from ..tol_colors import tol_cset


def fig_draw_lplot(axx, xdata, ydata, icol: str, use_steps=False, **kwargs):
    """
    add line plot to figure
    """
    # define colors
    cset = tol_cset('bright')
    color = cset[icol % len(cset)]

    if use_steps:
        edges = np.append(xdata, xdata[-1])
        values = np.append(ydata, ydata[-1])
        axx.stairs(values, edges, color=color, **kwargs)
    else:
        axx.plot(xdata, ydata, color=color, **kwargs)


def close_draw_lplot(axx, time_axis: bool, title: str, **kwargs):
    """
    close the figure created with MONplot::draw_lplot()
    """
    # add title to image panel
    if title is not None:
        axx.set_title(title, fontsize='large')
    # add grid lines (default settings)
    axx.grid(True)
    # add X & Y label
    if 'xlabel' in kwargs:
        axx.set_xlabel(kwargs['xlabel'])
    if 'ylabel' in kwargs:
        axx.set_ylabel(kwargs['ylabel'])
    # set the limits of the X-axis & Y-axis
    if 'xlim' in kwargs:
        axx.set_xlim(kwargs['xlim'])
    if 'ylim' in kwargs:
        axx.set_ylim(kwargs['ylim'])
    # set the scale of the X & Y axis {"linear", "log", "symlog", ...}
    if 'xscale' in kwargs:
        axx.set_xscale(kwargs['xscale'])
    if 'yscale' in kwargs:
        axx.set_ylabel(kwargs['yscale'])
    # format the X-axis when it is a time-axis
    if time_axis:
        plt.gcf().autofmt_xdate()
        plt.gca().xaxis.set_major_formatter(DateFormatter('%H:%M:%S'))

    # draw legenda in figure
    if axx.get_legend_handles_labels()[1]:
        axx.legend(fontsize='small', loc='best')



"""
This file is part of moniplot

https://github.com/rmvanhees/moniplot.git

This module

Functions
---------
fig_draw_tracks(axx, lons, lats, icids, saa_region)
   Add a subplot for tracks of the satellite projected on the Earth surface.

Copyright (c) 2022 SRON - Netherlands Institute for Space Research

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

License:  GPLv3
"""
import numpy as np

from cartopy import crs as ccrs

from matplotlib.patches import Polygon
import matplotlib.pyplot as plt

from ..tol_colors import tol_cset


# - main functions ---------------------------------
def fig_draw_tracks(axx, lons, lats, icids, saa_region) -> None:
    """
    Add a subplot for tracks of the satellite projected on the Earth surface
    """
    # define colors
    cset = tol_cset('bright')

    # draw coastlines and gridlines
    axx.set_global()
    axx.coastlines(resolution='110m')
    axx.gridlines()

    # draw satellite position(s)
    icolor = 0
    for val in np.unique(icids):
        mask = icids == val
        # pylint: disable=abstract-class-instantiated
        plt.plot(lons[mask], lats[mask], linestyle='',
                 marker='s', markersize=2,
                 color=cset[icolor % 6], label=f'ICID: {val}',
                 transform=ccrs.PlateCarree())
        icolor += 1
    axx.legend(loc='lower left')

    # draw SAA region
    if saa_region is not None:
        # pylint: disable=abstract-class-instantiated
        saa_poly = Polygon(xy=saa_region, closed=True, alpha=1.0,
                           facecolor=cset.grey, transform=ccrs.PlateCarree())
        axx.add_patch(saa_poly)
